/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "c_ppvyystup.h"
#include "cppkni.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct pp_vyypis {
    FILE *vyystup;
    const cpp_token *predoshlyy;
    const cpp_token *zdroj; /* Zdrojovyy token pre medzery. */
    int riadok;
    int vypiisanee; /* 1, ak niecsho bolo napiisanee na riadok. */
    const char *zdrojovyy_suubor;
};

static int daj_riadok(pp_vyypis *vs, lokaacia_t lok, const char *vlajky)
{
    if (vs->vypiisanee)
    {
        putc('\n', vs->vyystup);
        vs->vypiisanee = 0;
    }

    {
        const char *meno_suuboru = lok.suubor;
        size_t dllzhka_mena_suuboru = strlen(meno_suuboru);
        unsigned char *novee_meno = malloc(dllzhka_mena_suuboru * 4 + 1);
        unsigned char *s;

        if (novee_meno == NULL)
        {
            printf("Chyba pri alokaacii\n");
            abort();
        }

        vs->riadok = lok.riadok;
        vs->zdrojovyy_suubor = meno_suuboru;

        s = cpp_uuvodzovkuj_rethazec(novee_meno,
                                     (const unsigned char *) meno_suuboru,
                                     dllzhka_mena_suuboru);
        *s = '\0';

        fprintf(vs->vyystup, "# %u \"%s\"%s\n",
                vs->riadok == 0 ? 1 : vs->riadok,
                novee_meno, vlajky);

        free(novee_meno);
    }
    
    return (1);
}

static int mozhno_daj_riadok(pp_vyypis *vs, lokaacia_t lok)
{
    int umiestnenyy_oznacovacsh_riadka = 0;
    int orig_riadok = lok.riadok;
    const char *orig_suubor = lok.suubor;

    if (vs->vypiisanee)
    {
        putc('\n', vs->vyystup);
        vs->riadok++;
        vs->vypiisanee = 0;
    }

    if ((orig_riadok >= vs->riadok)
        && (orig_riadok < vs->riadok + 8)
        && (orig_suubor == vs->zdrojovyy_suubor))
    {
        while (orig_riadok > vs->riadok)
        {
            putc('\n', vs->vyystup);
            vs->riadok++;
        }
    }
    else umiestnenyy_oznacovacsh_riadka = daj_riadok(vs, lok, "");
    
    return (umiestnenyy_oznacovacsh_riadka);
}

static int urob_zmenu_riadka(pp_vyypis *vs, cpp_cshii *cshii,
                             const cpp_token *token, lokaacia_t lok,
                             int zbieranie_argumentov)
{
    int umiestnenyy_oznacovacsh_riadka = 0;
    
    if ((token->typ == CPP_KONIEC) || zbieranie_argumentov)
    {
        return (0);
    }

    umiestnenyy_oznacovacsh_riadka = mozhno_daj_riadok(vs, lok);
    vs->predoshlyy = NULL;
    vs->zdroj = NULL;

    {
        int medzery = lok.stllpec - 2;
        vs->vypiisanee = 1;

        while (--medzery >= 0)
        {
            putc(' ', vs->vyystup);
        }
    }
    
    return (umiestnenyy_oznacovacsh_riadka);
}

static void zmena_riadka(cpp_cshii *cshii, const cpp_token *token,
                         int zbieranie_argumentov)
{
    urob_zmenu_riadka(cpp_daj_pomoc2(cshii), cshii,
                      token, token->lokaacia,
                      zbieranie_argumentov);
}

static void zmena_suuboru(cpp_cshii *cshii,
                          lokaacia_t lok, int duovod)
{
    const char *vlajky = "";

    switch(duovod)
    {
        case CPP_D_VSTUP: vlajky = " 1"; break;
        case CPP_D_ODCHOD: vlajky = " 2"; break;
    }
    
    daj_riadok(cpp_daj_pomoc2(cshii), lok, vlajky);
}

pp_vyypis *inic_pp_vyystup(cpp_cshii *cshii, FILE *vyystup)
{
    pp_vyypis *vs = malloc(sizeof(*vs));

    if (vs == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    cpp_daj_spaetneevolania(cshii)->zmena_riadka = &zmena_riadka;
    cpp_daj_spaetneevolania(cshii)->zmena_suuboru = &zmena_suuboru;
    cpp_nastav_pomoc2(cshii, vs);

    vs->riadok = 1;
    vs->vypiisanee = 0;
    vs->predoshlyy = NULL;
    vs->zdrojovyy_suubor = "";
    vs->vyystup = vyystup;

    return (vs);
}

void preprocesuj_suubor(pp_vyypis *vs, cpp_cshii *cshii)
{
    int vyhni_sa_zlepeniu = 0;
    int umiestnenyy_oznacovacsh_riadka = 0;
    int rob_upraavy_riadkov = 1;

    vs->zdroj = NULL;
    for (;;)
    {
        lokaacia_t lok;
        const cpp_token *token = cpp_daj_token_s_lokaaciou(cshii, &lok);

        if (token->typ == CPP_KONIEC) break;
        
        if (token->typ == CPP_NICSH)
        {
            vyhni_sa_zlepeniu = 1;
            if ((vs->zdroj == NULL)
                || (!(vs->zdroj->vlajky & ZA_MEDZEROU)
                    && (token->hodnota.zdroj == NULL)))
            {
                vs->zdroj = token->hodnota.zdroj;
            }
            continue;
        }

        if (vyhni_sa_zlepeniu)
        {
            int orig_riadok = lok.riadok;

            if (vs->zdroj == NULL)
            {
                vs->zdroj = token;
            }

            if (rob_upraavy_riadkov
                && (orig_riadok != vs->riadok))
            {
                umiestnenyy_oznacovacsh_riadka = urob_zmenu_riadka(vs, cshii,
                                                                   token, lok,
                                                                   0);
                putc(' ', vs->vyystup);
                vs->vypiisanee = 1;
            }
            else if ((vs->zdroj->vlajky & ZA_MEDZEROU)
                     || (vs->predoshlyy
                         && cpp_vyhni_sa_zlepeniu(cshii,
                                                  vs->predoshlyy, token))
                     || ((vs->predoshlyy == NULL)
                         && (token->typ == CPP_MREZHA)))
            {
                putc(' ', vs->vyystup);
                vs->vypiisanee = 1;
            }
        }
        else if (token->vlajky & ZA_MEDZEROU)
        {
            int orig_riadok = lok.riadok;

            if (rob_upraavy_riadkov
                && (orig_riadok != vs->riadok))
            {
                umiestnenyy_oznacovacsh_riadka = urob_zmenu_riadka(vs, cshii,
                                                                   token, lok,
                                                                   0);
            }
                
            putc(' ', vs->vyystup);
            vs->vypiisanee = 1;
        }

        vyhni_sa_zlepeniu = 0;
        vs->zdroj = NULL;
        vs->predoshlyy = token;

        cpp_vypiish_token(token, vs->vyystup);
        umiestnenyy_oznacovacsh_riadka = 0;
        vs->vypiisanee = 1;
    }

    if (vs->vypiisanee) putc('\n', vs->vyystup);

    if (vs->vyystup != stdout) fclose(vs->vyystup);
    free(vs);
}
