/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "cppkni.h"
#include "internee.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum {
    PIISANIE_INTERPUNKCIA,
    PIISANIE_IDENT,
    PIISANIE_LITERAAL,
    PIISANIE_ZHIADNE
};

struct piisanie {
    int typ;
    unsigned char *text;
};

static const unsigned char *piisanie_digrafov[] = {
    (const unsigned char *)"%:",
    (const unsigned char *)"%:%:",
    (const unsigned char *)"<:",
    (const unsigned char *)":>",
    (const unsigned char *)"<%",
    (const unsigned char *)"%>"
};

#define JANEVIEM(a, b) {PIISANIE_INTERPUNKCIA, (unsigned char *)b},
#define JANEVIEM2(a, b) {PIISANIE_ ## b, NULL},
static const struct piisanie piisanie_tokenov[] = {TABULHKA_TYPOV};
#undef JANEVIEM
#undef JANEVIEM2

int _cpp_equiv_tokeny(const cpp_token *t1, const cpp_token *t2)
{
    if ((t1->typ == t2->typ) && (t1->vlajky == t2->vlajky))
    {
        switch (piisanie_tokenov[t1->typ].typ)
        {
            case PIISANIE_INTERPUNKCIA:
                return ((t1->typ != CPP_ZLEP)
                        || (t1->hodnota.cshiislo_tokenu
                            == t2->hodnota.cshiislo_tokenu));
            case PIISANIE_ZHIADNE:
                return ((t1->typ != CPP_MAKRO_ARGUMENT)
                        || ((t1->hodnota.makro_argument.cshiislo
                             == t2->hodnota.makro_argument.cshiislo)
                            && (t1->hodnota.makro_argument.piisanie
                                == t2->hodnota.makro_argument.piisanie)));
            case PIISANIE_IDENT:
                return (t1->hodnota.neviem == t2->hodnota.neviem);
            case PIISANIE_LITERAAL:
                return ((t1->hodnota.rethazec.dllzhka
                         == t2->hodnota.rethazec.dllzhka)
                        && (memcmp(t1->hodnota.rethazec.text,
                                   t2->hodnota.rethazec.text,
                                   t1->hodnota.rethazec.dllzhka) == 0));
        }
    }
            
    return (0);
}

struct cpp_spaetneevolania *cpp_daj_spaetneevolania(cpp_cshii *cshii)
{
    return (&(cshii->spaetneevolania));
}

static cpp_neviem2 *novee_neviem2(cpp_cshii *cshii)
{
    cpp_neviem2 *novee = malloc(sizeof(*novee));

    if (novee == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    novee->dhalshie = cshii->neviem2;
    cshii->neviem2 = novee;

    return (novee);
}

static cpp_neviem *makro_neviem2(cpp_neviem2 *neviem2)
{
    if (neviem2 == NULL) return (NULL);

    return (neviem2->makro);
}

void _cpp_pridaj_neviem2_priame(cpp_cshii *cshii, cpp_neviem *neviem,
                                const cpp_token *prvyy, int pocshet)
{
    cpp_neviem2 *neviem2 = novee_neviem2(cshii);

    if (neviem == NULL)
    {
        neviem = makro_neviem2(cshii->neviem2->dhalshie);
    }

    neviem2->typ_tokenov = TOKENY_TYP_PRIAME;
    neviem2->makro = neviem;
    neviem2->pamaeth = NULL;
    neviem2->prvyy.token = prvyy;
    neviem2->poslednyy.token = prvyy + pocshet;
}

static void pridaj_neviem2_nepriame(cpp_cshii *cshii, cpp_neviem *neviem,
                                    void *pamaeth,
                                    const cpp_token **stokeny,
                                    unsigned int pocshet)
{
    cpp_neviem2 *neviem2 = novee_neviem2(cshii);

    if (neviem == NULL)
    {
        neviem = makro_neviem2(cshii->neviem2->dhalshie);
    }

    neviem2->typ_tokenov = TOKENY_TYP_NEPRIAME;
    neviem2->makro = neviem;
    neviem2->pamaeth = pamaeth;
    neviem2->prvyy.stoken = stokeny;
    neviem2->poslednyy.stoken = stokeny + pocshet;
}

void _cpp_odstraanh_neviem2(cpp_cshii *cshii)
{
    cpp_neviem2 *neviem2 = cshii->neviem2;

    /* Je NULL, ak bolo neviem2 vytvorenee expanduj_arg(). */
    if (neviem2->makro)
    {
        cpp_neviem *makro = neviem2->makro;

        if ((makro != NULL)
            && (makro_neviem2(neviem2->dhalshie) != makro))
            makro->vlajky &= ~NEVIEM_VYPNUTEE;
    }

    free(neviem2->pamaeth);

    cshii->neviem2 = neviem2->dhalshie;
    
    free(neviem2);
}

void _cpp_vraath_tokeny_lexeru(cpp_cshii *cshii, unsigned int pocshet)
{
    cshii->tokeny_vpredu += pocshet;
    /* _cpp_lex_token() nikdy nenechaa
     * cshii->akt_token == cshii->akt_tokenrad->shtart.*/
    while (pocshet--)
    {
        cshii->akt_token--;
        if ((cshii->akt_token == cshii->akt_tokenrad->shtart)
            && (cshii->akt_tokenrad->predoshlyy != NULL))
        {
            cshii->akt_tokenrad = cshii->akt_tokenrad->predoshlyy;
            cshii->akt_token = cshii->akt_tokenrad->koniec;
        }
    }
}

/* Vraati spaeth tokeny urcshenee pocshtom, pre tokeny z makier
 * dokaazhe vraatith len 1. */
void _cpp_vraath_tokeny(cpp_cshii *cshii, unsigned int pocshet)
{
    if (cshii->neviem2 == NULL)
    {
        _cpp_vraath_tokeny_lexeru(cshii, pocshet);
    }
    else
    {
        if (pocshet != 1) abort();
        switch (cshii->neviem2->typ_tokenov)
        {
            case TOKENY_TYP_PRIAME:
                cshii->neviem2->prvyy.token--;
                break;
            case TOKENY_TYP_NEPRIAME:
                cshii->neviem2->prvyy.stoken--;
                break;
        }
    }
}

static cpp_token *nicsh_token(cpp_cshii *cshii, const cpp_token *pred_token)
{
    cpp_token *vyysledok = _cpp_docshasnyy_token(cshii);

    vyysledok->typ = CPP_NICSH;
    vyysledok->hodnota.zdroj = pred_token;
    vyysledok->vlajky = 0;

    return (vyysledok);
}

enum makro_arg_token_typ {
    MAKRO_ARG_TOKEN_OBYCSHAJNYY,
    MAKRO_ARG_TOKEN_EXPANDOVANYY
};

typedef struct {
    const cpp_token **prvyy; /* Prvyy neexpandovanyy token. */
    const cpp_token **expandovanyy;
    const cpp_token *zrethazcovanyy;
    unsigned int pocshet;
    unsigned int expandovanyy_pocshet;
} makro_arg;

static const cpp_token **arg_token_smerniik(makro_arg *arg,
                                             unsigned int index,
                                             enum makro_arg_token_typ typ)
{
    const cpp_token **token_s;

    switch (typ)
    {
        case MAKRO_ARG_TOKEN_OBYCSHAJNYY: token_s = arg->prvyy; break;
        case MAKRO_ARG_TOKEN_EXPANDOVANYY: token_s = arg->expandovanyy; break;
    }

    return (&(token_s[index]));
}

static void pridaj_arg_token(makro_arg *arg, const cpp_token *token,
                             unsigned int index,
                             enum makro_arg_token_typ typ)
{
    const cpp_token **token_s;

    token_s = arg_token_smerniik(arg, index, typ);
    *token_s = token;
}

int _cpp_argumenty_v_poriadku(cpp_cshii *cshii,
                              cpp_makro *makro,
                              cpp_neviem *neviem,
                              unsigned int pocshet_argumentov)
{
    if (pocshet_argumentov == makro->pocshet_parametrov)
    {
        return (1);
    }
    if (pocshet_argumentov < makro->pocshet_parametrov)
    {
        if ((pocshet_argumentov + 1 < makro->pocshet_parametrov)
            && makro->variadickee)
        {
            cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                      "ISO C17 vyzhaduje asponh jeden argument pre \"...\""
                      " vo variadickom makre");
            return (1);
        }

        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "makro \"%s\" potrebuje %u argumenty, ale dostalo len %u",
                  NEVIEM_MENO(neviem), makro->pocshet_parametrov,
                  pocshet_argumentov);
    }
    else
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "makro \"%s\" berie len %u argumenty, ale dostalo %u",
                  NEVIEM_MENO(neviem), makro->pocshet_parametrov,
                  pocshet_argumentov);
    }

    return (0);
}

static void *zbieraj_argumenty(cpp_cshii *cshii, cpp_neviem *neviem,
                               unsigned int *pocshet_argumentov_s)
{
    cpp_makro *makro = neviem->hodnota.makro;
    unsigned char *pamaeth;
    makro_arg *argy, *arg;
    const cpp_token **token_s;
    unsigned int pocshet_parametrov;
    size_t velhkosth;
    const cpp_token *token;
    unsigned int pocshet_argumentov;

#define TOKENY_NA_ARGUMENT 17

    if (makro->pocshet_parametrov) pocshet_parametrov = makro->pocshet_parametrov;
    else pocshet_parametrov = 1;

    velhkosth = ((sizeof(makro_arg) + sizeof(cpp_token *) * TOKENY_NA_ARGUMENT)
                * pocshet_parametrov);
    pamaeth = malloc(velhkosth);
    if (pamaeth == NULL)
    {
        printf("Chyba pri alokaacii pamaete\n");
        abort();
    }
    memset(pamaeth, '\0', velhkosth);

    argy = (makro_arg *)pamaeth;
    token_s = (const cpp_token **)(argy + pocshet_parametrov);

    arg = argy;
    pocshet_argumentov = 0;

    do {
        unsigned int hllbka_zaatvoriek = 0;
        unsigned int pocshet_tokenov = 0;
        
        pocshet_argumentov++;
        arg->prvyy = token_s;

        for (;;)
        {
            if (&(arg->prvyy[pocshet_tokenov + 2])
                > (const cpp_token **)(pamaeth + velhkosth))
            {
                unsigned char *staryy = pamaeth;

                velhkosth += 987 * sizeof(cpp_token *);
                pamaeth = realloc(pamaeth, velhkosth);
                if (pamaeth == NULL)
                {
                    printf("Chyba pri alokaacii pamaete\n");
                    abort();
                }

                argy = (makro_arg *)pamaeth;
                arg = argy + (pocshet_argumentov - 1);
                token_s = (const cpp_token **)(pamaeth
                                               + (((unsigned char *)token_s)
                                                  - staryy));
                arg->prvyy = token_s;
            }

            token = cpp_daj_token(cshii);

            if (token->typ == CPP_NICSH)
            {
                /* Zahodii medzery na zacshiatku. */
                if (pocshet_tokenov == 0) continue;
            }
            else if (token->typ == CPP_LH_ZAATVORKA) hllbka_zaatvoriek++;
            else if (token->typ == CPP_P_ZAATVORKA)
            {
                if (hllbka_zaatvoriek) hllbka_zaatvoriek--;
                else break;
            }
            else if (token->typ == CPP_CSHIARKA)
            {
                if (hllbka_zaatvoriek == 0
                    && !(makro->variadickee
                         && (pocshet_argumentov
                             == makro->pocshet_parametrov))) break;
            }
            else if (token->typ == CPP_KONIEC)
            {
                break;
            }

            pridaj_arg_token(arg, token,
                             pocshet_tokenov, MAKRO_ARG_TOKEN_OBYCSHAJNYY);
            pocshet_tokenov++;
        }

        /* Zahodii medzery na konci. */
        while (pocshet_tokenov > 0
               && (arg->prvyy[pocshet_tokenov - 1]->typ == CPP_NICSH))
        {
            pocshet_tokenov--;
        }

        arg->pocshet = pocshet_tokenov;
        pridaj_arg_token(arg, &(cshii->koniec),
                         pocshet_tokenov, MAKRO_ARG_TOKEN_OBYCSHAJNYY);

        if (pocshet_argumentov <= makro->pocshet_parametrov)
        {
            token_s = &(arg->prvyy[pocshet_tokenov + 1]);
            if (pocshet_argumentov != makro->pocshet_parametrov)
            {
                arg++;
            }
        }
    } while ((token->typ != CPP_P_ZAATVORKA) && (token->typ != CPP_KONIEC));

    if (token->typ == CPP_KONIEC)
    {
        if (cshii->neviem2 || cshii->stav.v_direktiive)
        {
            _cpp_vraath_tokeny(cshii, 1);
        }
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "neukoncshenyy zoznam argumentov pre makro \"%s\"",
                  NEVIEM_MENO(neviem));
    }
    else
    {
        /* Jeden praazdny argument je nula argumentov. */
        if ((pocshet_argumentov == 0)
            && (makro->pocshet_parametrov == 0)
            && (argy[0].pocshet == 0))
        {
            pocshet_argumentov = 0;
        }
        if (_cpp_argumenty_v_poriadku(cshii, makro, neviem, pocshet_argumentov))
        {
            if (pocshet_argumentov_s) *pocshet_argumentov_s = pocshet_argumentov;

            return (pamaeth);
        }
    }

    free(pamaeth);
    
    return NULL;
}

static void *funkciovee_neviemcsho(cpp_cshii *cshii, cpp_neviem *neviem,
                                   unsigned int *pocshet_argumentov)
{
    const cpp_token *token, *nicsh = NULL;

    for (;;)
    {
        token = cpp_daj_token(cshii);
        if (token->typ != CPP_NICSH) break;
        if ((nicsh == NULL)
            || (!(nicsh->vlajky & ZA_MEDZEROU)
                && (token->hodnota.zdroj == NULL)))
        {
            nicsh = token;
        }
    }

    if (token->typ == CPP_LH_ZAATVORKA)
    {
        cshii->stav.zbieranie_argumentov = 2;
        return (zbieraj_argumenty(cshii, neviem, pocshet_argumentov));
    }

    if ((token->typ != CPP_KONIEC) || (token == &(cshii->koniec)))
    {
        _cpp_vraath_tokeny(cshii, 1);
        if (nicsh)
        {
            _cpp_pridaj_neviem2_priame(cshii, NULL, nicsh, 1);
        }
    }

    return NULL;
}

static void vymazh_argumenty(void *pamaeth, unsigned int pocshet_argumentov)
{
    makro_arg *argy = pamaeth;
    unsigned int i;

    if (pamaeth == NULL) return;

    for (i = 0; i < pocshet_argumentov; i++)
    {
        free(argy[i].expandovanyy);
        argy[i].expandovanyy = NULL;
    }

    free(pamaeth);
}

static unsigned int makro_skutocshnyy_pocshet_tokenov(cpp_makro *makro)
{
    unsigned int i;
    
    if (makro->tokeny_naviac == 0) return (makro->pocshet);

    for (i = makro->pocshet; i--;)
    {
        if (makro->tokeny[i].typ != CPP_ZLEP)
            return (i + 1);
    }

    return (0);
}

/* Alokuje a vraathi CPP_RETHAZEC token s TEXTom DLLZHKy.
 * TEXT musii byth permanentne alokovanyy. */
static const cpp_token *novyy_rethazcovyy_token(cpp_cshii *cshii,
                                                unsigned char *text,
                                                unsigned int dllzhka)
{
    cpp_token *token = _cpp_docshasnyy_token(cshii);

    text[dllzhka] = '\0';
    token->typ = CPP_RETHAZEC;
    token->hodnota.rethazec.text = text;
    token->hodnota.rethazec.dllzhka = dllzhka;
    token->vlajky = 0;

    return (token);
}

static const cpp_token *zrethazcuj_arg(cpp_cshii *cshii, makro_arg *arg)
{
    unsigned char *pamaeth, *koniec, *cielh;
    unsigned int i, unikni_to, pocshet_spaetnyychlomiek = 0;
    const cpp_token *zdroj = NULL;
    size_t dllzhka;

    cielh = pamaeth = malloc(3);
    if (pamaeth == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }
    koniec = pamaeth + 3;
    *(cielh++) = '"';

    for (i = 0; i < arg->pocshet; i++)
    {
        const cpp_token *token = arg->prvyy[i];

        if (token->typ == CPP_NICSH)
        {
            if ((zdroj == NULL)
                || (!(zdroj->vlajky & ZA_MEDZEROU)
                    && (token->hodnota.zdroj == NULL)));
            {
                zdroj = token->hodnota.zdroj;
            }
            continue;
        }

        unikni_to = ((token->typ == CPP_RETHAZEC)
                     || (token->typ == CPP_LRETHAZEC)
                     || (token->typ == CPP_uRETHAZEC)
                     || (token->typ == CPP_URETHAZEC)
                     || (token->typ == CPP_u8RETHAZEC)
                     || (token->typ == CPP_ZNAK)
                     || (token->typ == CPP_LZNAK)
                     || (token->typ == CPP_uZNAK)
                     || (token->typ == CPP_UZNAK));

        dllzhka = cpp_token_dllzhka(token);
        if (unikni_to) dllzhka *= 4;
        dllzhka += 3;

        if (((size_t)(koniec - cielh)) < dllzhka)
        {
            size_t dllzhka_doteraz = cielh - pamaeth;
            size_t velhkosth = koniec - pamaeth + dllzhka;
            pamaeth = realloc(pamaeth, velhkosth);
            if (pamaeth == NULL)
            {
                printf("Chyba pri alokaacii\n");
                abort();
            }
            koniec = pamaeth + velhkosth;
            cielh = pamaeth + dllzhka_doteraz;
        }

        if (cielh - 1 != pamaeth)
        {
            if (zdroj == NULL) zdroj = token;
            if (zdroj->vlajky & ZA_MEDZEROU) *(cielh++) = ' ';
        }
        zdroj = NULL;

        if (unikni_to)
        {
            unsigned char *ppamaeth = malloc(dllzhka);
            if (ppamaeth == NULL)
            {
                printf("Chyba pri alokaacii\n");
                abort();
            }
            dllzhka = cpp_napiish_token(cshii, token, ppamaeth, 1) - ppamaeth;
            cielh = cpp_uuvodzovkuj_rethazec(cielh, ppamaeth, dllzhka);
            free(ppamaeth);
        }
        else cielh = cpp_napiish_token(cshii, token, cielh, 1);

        if (token->typ == CPP_INEE && token->hodnota.rethazec.text[0] == '\\')
            pocshet_spaetnyychlomiek++;
        else pocshet_spaetnyychlomiek = 0;
    }

    if (pocshet_spaetnyychlomiek & 1)
    {
        cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                  "nespraavny rethazcovyy literaal, ignoruje sa poslednaa '\\'");
        cielh--;
    }

    *(cielh++) = '"';
    dllzhka = cielh - pamaeth;
    return (novyy_rethazcovyy_token(cshii, cielh - dllzhka, dllzhka));
}

static void expanduj_arg(cpp_cshii *cshii, makro_arg *arg)
{
    size_t velhkosth;
    
    if (arg->pocshet == 0 || arg->expandovanyy != NULL) return;

    velhkosth = 356;
    arg->expandovanyy = malloc(sizeof(const cpp_token *) * velhkosth);

    if (arg->expandovanyy == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    pridaj_neviem2_nepriame(cshii, NULL, NULL,
                            arg->prvyy, arg->pocshet + 1);

    for (;;)
    {
        const cpp_token *token;

        if (arg->expandovanyy_pocshet + 1 > velhkosth)
        {
            velhkosth *= 2;
            arg->expandovanyy = realloc(arg->expandovanyy, velhkosth);
            if (arg->expandovanyy == NULL)
            {
                printf("Chyba pri alokaacii\n");
                abort();
            }
        }

        token = cpp_daj_token(cshii);

        if (token->typ == CPP_KONIEC) break;

        pridaj_arg_token(arg, token,
                         arg->expandovanyy_pocshet,
                         MAKRO_ARG_TOKEN_EXPANDOVANYY);
        arg->expandovanyy_pocshet++;
    }

    _cpp_odstraanh_neviem2(cshii);
}

static void *token_pamaeth_novaa(cpp_cshii *cshii, size_t velhkosth)
{
    void *pamaeth = malloc(sizeof(cpp_token *) * velhkosth);

    if (pamaeth == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    return (pamaeth);
}

static void nahradh_argumenty(cpp_cshii *cshii, cpp_neviem *neviem,
                              cpp_makro *makro, makro_arg *argy)
{
    unsigned int celkom_tokenov, neviem_cshiislo, i;
    const cpp_token *token, *koniec;
    makro_arg *arg;
    void *pamaeth;
    const cpp_token **prvyy;

    neviem_cshiislo = makro_skutocshnyy_pocshet_tokenov(makro);
    celkom_tokenov = neviem_cshiislo;
    koniec = makro->tokeny + neviem_cshiislo;

    for (token = makro->tokeny; token < koniec; token++)
    {
        if (token->typ == CPP_MAKRO_ARGUMENT)
        {
            /* CPP_NICSH pred a za. */
            celkom_tokenov += 2;

            arg = argy + token->hodnota.makro_argument.cshiislo;

            if (token->vlajky & ZRETHAZCUJ_ARGUMENT)
            {
                if (!(arg->zrethazcovanyy))
                    arg->zrethazcovanyy = zrethazcuj_arg(cshii, arg);
            }
            else if ((token->vlajky & ZLEP_LHAVO)
                     || (token != makro->tokeny
                         && token[-1].vlajky & ZLEP_LHAVO))
            {
                celkom_tokenov += arg->pocshet - 1;
            }
            else
            {
                if (!(arg->expandovanyy)) expanduj_arg(cshii, arg);
                celkom_tokenov += arg->expandovanyy_pocshet - 1;
            }
        }
    }

    pamaeth = token_pamaeth_novaa(cshii, celkom_tokenov);
    prvyy = (const cpp_token **)pamaeth;

    i = 0;
    for (token = makro->tokeny; token < koniec; token++)
    {
        unsigned int arg_pocshet_tokenov;
        enum {
            RETHAZEC, NEEXPAND, EXPAND
            } typ_argu;
        
        if (token->typ != CPP_MAKRO_ARGUMENT)
        {
            *(prvyy++) = token;
            i++;
            continue;
        }

        arg = argy + token->hodnota.makro_argument.cshiislo;

        if (token->vlajky & ZRETHAZCUJ_ARGUMENT)
        {
            arg_pocshet_tokenov = 1;
            typ_argu = RETHAZEC;
        }
        else if (token->vlajky & ZLEP_LHAVO)
        {
            arg_pocshet_tokenov = arg->pocshet;
            typ_argu = NEEXPAND;
        }
        else if ((token != makro->tokeny) && (token[-1].vlajky & ZLEP_LHAVO))
        {
            arg_pocshet_tokenov = arg->pocshet;
            typ_argu = NEEXPAND;
        }
        else
        {
            arg_pocshet_tokenov = arg->expandovanyy_pocshet;
            typ_argu = EXPAND;
        }

        if ((token != makro->tokeny) && (!(token[-1].vlajky & ZLEP_LHAVO)))
        {
            const cpp_token *m = nicsh_token(cshii, token);

            *(prvyy++) = m;
        }

        if (arg_pocshet_tokenov)
        {
            unsigned int j;
            for (j = 0; j < arg_pocshet_tokenov; j++)
            {
                const cpp_token *t;
                switch (typ_argu)
                {
                    case RETHAZEC: t = arg->zrethazcovanyy; break;
                    case NEEXPAND: t = arg->prvyy[j]; break;
                    case EXPAND: t = arg->expandovanyy[j]; break;
                }

                *(prvyy++) = t;
            }
        }
        else
        {
            /* +++VAROVANIA PRE C90 */;
        }

        if (!(cshii->stav.v_direktiive) && !(token->vlajky & ZLEP_LHAVO))
        {
            const cpp_token *m = &(cshii->blokuj_zlepenie);
            
            *(prvyy++) = m;
        }

        i += arg_pocshet_tokenov;
    }

    pridaj_neviem2_nepriame(cshii, neviem, pamaeth,
                            (const cpp_token **)pamaeth,
                            prvyy - (const cpp_token **)pamaeth);
}

#include <time.h>

unsigned char *_cpp_vstavanee_makro_text(cpp_cshii *cshii,
                                         cpp_neviem *neviem,
                                         lokaacia_t lok)
{
    unsigned char *vyysledok = NULL;
    unsigned long cshiislo = 1;

    switch (neviem->hodnota.vstavanee)
    {
        default:
            cpp_chyba(cshii, CPP_DUU_INTERNAA_CHYBA,
                      "nespravnee vstavanee makro \"%s\"",
                      NEVIEM_MENO(neviem));
            break;

        case VS_DATE:
        case VS_TIME:
            if (cshii->daatum == NULL)
            {
                time_t tt = time(NULL);
                struct tm *ts = NULL;

                errno = 0;
                if ((tt != (time_t)-1) || (errno == 0))
                {
                    ts = localtime(&tt);
                }

                cshii->daatum = malloc(sizeof("\"Abc 12 3456\""));
                cshii->cshas = malloc(sizeof("\"12:34:56\""));
                if ((cshii->daatum == NULL)
                    || (cshii->cshas == NULL))
                {
                    printf("Chyba pri alokacii\n");
                    abort();
                }

                if (ts)
                {
                    char *mesiace[] = {
                        "Jan", "Feb", "Mar", "Apr", "Maj", "Jun",
                        "Jul", "Aug", "Sep", "Okt", "Nov", "Dec"};
                    sprintf((char *)(cshii->daatum), "\"%s %2d %4d\"",
                            mesiace[ts->tm_mon], ts->tm_mday,
                            ts->tm_year + 1900);
                    sprintf((char *)(cshii->cshas), "\"%02d:%02d:%02d\"",
                            ts->tm_hour, ts->tm_min, ts->tm_sec);
                }
                else
                {
                    cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                              "nepodarilo sa zisthith daatum a cshas");
                    strcpy(cshii->daatum, "\"??? ?? ????\"");
                    strcpy(cshii->cshas, "\"??:??:??\"");
                }
            }

            if (neviem->hodnota.vstavanee == VS_DATE)
            {
                vyysledok = malloc(strlen(cshii->daatum) + 1);
                if (vyysledok == NULL)
                {
                    printf("Chyba pri alokacii\n");
                    abort();
                }
                memcpy(vyysledok, cshii->daatum, strlen(cshii->daatum) + 1);
            }
            else
            {
                vyysledok = malloc(strlen(cshii->cshas) + 1);
                if (vyysledok == NULL)
                {
                    printf("Chyba pri alokacii\n");
                    abort();
                }
                memcpy(vyysledok, cshii->cshas, strlen(cshii->cshas) + 1);
            }
            break;

        case VS_FILE: {
            unsigned int dllzhka;
            const char *meno;
            unsigned char *p;

            meno = lok.suubor;

            dllzhka = strlen(meno);
            p = malloc(dllzhka * 2 + 3);
            if (p == NULL)
            {
                printf("Chyba pri alokaacii\n");
            }
            vyysledok = p;
            *p = '"';
            p = cpp_uuvodzovkuj_rethazec(p + 1, (const unsigned char *)meno,
                                         dllzhka);
            *(p++) = '"';
            *p = '\0';
            break;
        }

        case VS_LINE:
            cshiislo = lok.riadok;
            break;
    }

    if (vyysledok == NULL)
    {
        vyysledok = malloc(21);
        if (vyysledok == NULL)
        {
            printf("Chyba pri alokacii\n");
            abort();
        }

        sprintf((char *)vyysledok, "%lu", cshiislo);
    }

    return (vyysledok);
}

static int vstavanee_makro(cpp_cshii *cshii, cpp_neviem *neviem,
                           lokaacia_t lok, lokaacia_t expanznaa_lok)
{
    unsigned char *p;
    size_t dllzhka;
    unsigned char *p2;
    cpp_token *token;

    if (neviem->hodnota.vstavanee == VS_PRAGMA)
    {
        /* Neinterpretuje sa _Pragma v direktiivach. */
        if (cshii->stav.v_direktiive) return (0);

        printf("+++_Pragma nie je podporovanee\n");

        return (0);
    }

    p = _cpp_vstavanee_makro_text(cshii, neviem, expanznaa_lok);
    dllzhka = strlen((const char *)p);
    p2 = malloc(dllzhka + 1);
    if (p2 == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    memcpy(p2, p, dllzhka);
    free(p);
    p2[dllzhka] = '\n';

    _cpp_pridaj_ppos(cshii, p2, dllzhka, /* od_faazy3 */ 1);
    _cpp_cshisti_riadok(cshii);

    cshii->akt_token = _cpp_docshasnyy_token(cshii);
    token = _cpp_lex_token_j(cshii);
    token->lokaacia = lok;

    _cpp_pridaj_neviem2_priame(cshii, NULL, token, 1);

    if (cshii->ppos->poz != cshii->ppos->koniec)
    {
        cpp_chyba(cshii, CPP_DUU_INTERNAA_CHYBA,
                  "nespravnee vstavanee makro \"%s\"",
                  NEVIEM_MENO(neviem));
    }

    _cpp_odstraanh_ppos(cshii);
    
    return (1);
}

static int expanduj_makro(cpp_cshii *cshii, cpp_neviem *neviem,
                           const cpp_token *vyysledok, lokaacia_t lok)
{
    cshii->ide_sa_expandovath_makro = 1;
    if (neviem->typ == NEVIEM_MAKRO)
    {
        cpp_makro *makro = neviem->hodnota.makro;

        if (makro->funkciovee)
        {
            unsigned int pocshet_argumentov = 0;
            void *v;
            
            cshii->stav.blokuj_expanziu++;
            cshii->drzh_tokeny++;
            cshii->stav.zbieranie_argumentov = 1;
            v = funkciovee_neviemcsho(cshii, neviem, &pocshet_argumentov);
            cshii->stav.zbieranie_argumentov = 0;
            cshii->drzh_tokeny--;
            cshii->stav.blokuj_expanziu--;

            if (v == NULL)
            {
                cshii->ide_sa_expandovath_makro = 0;
                return (0);
            }

            if (makro->pocshet_parametrov > 0)
            {
                nahradh_argumenty(cshii, neviem, makro,
                                  (makro_arg *) v);
            }

            vymazh_argumenty(v, pocshet_argumentov);
        }

        neviem->vlajky |= NEVIEM_VYPNUTEE;

        if (makro->pocshet_parametrov == 0)
        {
            unsigned int pocshet_tokenov;
            pocshet_tokenov = makro_skutocshnyy_pocshet_tokenov(makro);
            _cpp_pridaj_neviem2_priame(cshii, neviem,
                                       makro->tokeny,
                                       pocshet_tokenov);
        }

        cshii->ide_sa_expandovath_makro = 0;
        return (1);
    }
    
    cshii->ide_sa_expandovath_makro = 0;
    {
        lokaacia_t expanznaa_lok = cshii->invokacshnaa_lokaacia;

        return (vstavanee_makro(cshii, neviem, lok, expanznaa_lok));
    }
}

static int koniec_neviem2(cpp_neviem2 *neviem2)
{
    switch (neviem2->typ_tokenov)
    {
        case TOKENY_TYP_PRIAME:
            return (neviem2->prvyy.token == neviem2->poslednyy.token);
        case TOKENY_TYP_NEPRIAME:
            return (neviem2->prvyy.stoken == neviem2->poslednyy.stoken);
    }
}

static const cpp_token *ber_token_z_neviem2(cpp_cshii *cshii)
{
    cpp_neviem2 *neviem2 = cshii->neviem2;

    switch (neviem2->typ_tokenov)
    {
        case TOKENY_TYP_PRIAME:
            return (neviem2->prvyy.token++);
        case TOKENY_TYP_NEPRIAME:
            return (*(neviem2->prvyy.stoken++));
    }

    printf("Vnuutornee zlyhanie\n");
    abort();
}

/* Skuusi zlepith 2 tokeny. Vraati 0 pri uuspechu.
 * slhavyy je zmenenee na token bez ZLPE_LHAVO. */
static int zlep_tokeny(cpp_cshii *cshii, const cpp_token **slhavyy,
                       const cpp_token *pravyy)
{
    unsigned char *pamaeth, *koniec, *lhavyy_koniec;
    cpp_token *lhavyy;
    unsigned int dllzhka;

    dllzhka = cpp_token_dllzhka(*slhavyy) + cpp_token_dllzhka(pravyy) + 1;
    pamaeth = malloc(dllzhka);
    if (pamaeth == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }
    koniec = lhavyy_koniec = cpp_napiish_token(cshii, *slhavyy, pamaeth, 1);

    /* Vyhne sa komentaarom.
     * Je jednochushie pridath medzeru nezh menith _cpp_lex_token_j(). */
    if (((*slhavyy)->typ == CPP_LOMKA) && (pravyy->typ != CPP_ROVNAA_SA))
        *(koniec++) = ' ';
    /* Zvlaashtny priipad, kedh praavy je CPP_NICSH. */
    if (pravyy->typ != CPP_NICSH)
        koniec = cpp_napiish_token(cshii, pravyy, koniec, 1);
    *koniec = '\n';

    _cpp_pridaj_ppos(cshii, pamaeth, koniec - pamaeth, /* od_faazy3 */ 1);
    _cpp_cshisti_riadok(cshii);

    cshii->akt_token = _cpp_docshasnyy_token(cshii);
    lhavyy = _cpp_lex_token_j(cshii);
    if (cshii->ppos->poz != cshii->ppos->koniec)
    {
        _cpp_odstraanh_ppos(cshii);
        _cpp_vraath_tokeny(cshii, 1);
        *lhavyy_koniec = '\0';

        *lhavyy = **slhavyy;
        *slhavyy = lhavyy;
        lhavyy->vlajky &= ~ZLEP_LHAVO;

        cpp_chyba(cshii, CPP_DUU_CHYBA, "zlepenie \"%s\" a \"%s\""
                  " nedaava platnyy preprocesorovyy token",
                  pamaeth, cpp_token_ako_text(pravyy));
        free(pamaeth);
        return (1);
    }

    free(pamaeth);
    *slhavyy = lhavyy;
    _cpp_odstraanh_ppos(cshii);
    return (0);
}

static void zlep_vshetky_tokeny(cpp_cshii *cshii, const cpp_token *lhavyy)
{
    const cpp_token *pravyy;
    cpp_neviem2 *neviem2 = cshii->neviem2;

    if ((neviem2 == NULL) || (!(lhavyy->vlajky & ZLEP_LHAVO)))
    {
        printf("Vnuutornee zlyhanie\n");
        abort();
    }

    do {
        /* Vezme sa token z aktuaalneho neviem2. */
        switch (neviem2->typ_tokenov)
        {
            case TOKENY_TYP_PRIAME:
                pravyy = neviem2->prvyy.token++; break;
            case TOKENY_TYP_NEPRIAME:
                pravyy = *(neviem2->prvyy.stoken++); break;
        }

        if (pravyy->typ == CPP_NICSH)
        {
            if (pravyy->vlajky & ZLEP_LHAVO)
            {
                printf("Vnuutornee zlyhanie\n");
                abort();
            }
        }
        if (zlep_tokeny(cshii, &lhavyy, pravyy)) break;
    } while (pravyy->vlajky & ZLEP_LHAVO);

    _cpp_pridaj_neviem2_priame(cshii, NULL, lhavyy, 1);
}

static int v_makro_expanzii(cpp_cshii *cshii)
{
    if (cshii == NULL) return (0);

    return (cshii->ide_sa_expandovath_makro
            || makro_neviem2(cshii->neviem2));
}

const cpp_token *cpp_daj_token_1(cpp_cshii *cshii,
                                 lokaacia_t *vlok)
{
    const cpp_token *vyysledok;
    lokaacia_t lok;

    for (;;)
    {
        cpp_neviem *neviem;
        cpp_neviem2 *neviem2 = cshii->neviem2;

        if (neviem2 == NULL)
        {
            vyysledok = _cpp_lex_token(cshii);
        }
        else if (!koniec_neviem2(neviem2))
        {
            vyysledok = ber_token_z_neviem2(cshii);
            if (vyysledok->vlajky & ZLEP_LHAVO)
            {
                zlep_vshetky_tokeny(cshii, vyysledok);
                if (cshii->stav.v_direktiive) continue;
                vyysledok = nicsh_token(cshii, vyysledok);
                break;
            }
        }
        else
        {
            _cpp_odstraanh_neviem2(cshii);
            
            if (cshii->stav.v_direktiive) continue;

            vyysledok = &(cshii->blokuj_zlepenie);
            break;
        }

        lok = vyysledok->lokaacia;

        if (vyysledok->vlajky & NEEXPANDUJ)
        {
            break;
        }

        if (vyysledok->typ != CPP_IDENT)
        {
            break;
        }

        neviem = vyysledok->hodnota.neviem;

        if (neviem->typ == NEVIEM_NICSH)
        {
            break;
        }

        if (!(neviem->vlajky & NEVIEM_VYPNUTEE))
        {
            int expandovanee = 0;

            if (!v_makro_expanzii(cshii))
            {
                cshii->invokacshnaa_lokaacia = lok;
            }

            if (cshii->stav.blokuj_expanziu)
            {
                break;
            }

            expandovanee = expanduj_makro(cshii, neviem, vyysledok,
                                          lok);

            if (expandovanee)
            {
                vyysledok = nicsh_token(cshii, vyysledok);
                break;
            }
        }
        else
        {
            /* Oznacshii tento token ako neexpandovatelhnyy. */
            cpp_token *t = _cpp_docshasnyy_token(cshii);
            t->typ = vyysledok->typ;
            t->vlajky = vyysledok->vlajky | NEEXPANDUJ;
            t->hodnota = vyysledok->hodnota;
            vyysledok = t;
        }

        break;
    }

    if (vlok)
    {
        *vlok = lok;

        if (makro_neviem2(cshii->neviem2) != NULL)
        {
            *vlok = cshii->invokacshnaa_lokaacia;
        }
    }
    
    return (vyysledok);
}

const cpp_token *cpp_daj_token(cpp_cshii *cshii)
{
    return (cpp_daj_token_1(cshii, NULL));
}

/* Vraathi token s lokaaciou mozhno odlishnou od token->lokaacia.
 * Ak token pochaadza z expanzie makra, vraati sa lokaacia zacshiatku
 * expanzie. */
const cpp_token *cpp_daj_token_s_lokaaciou(cpp_cshii *cshii, lokaacia_t *lok)
{
    return (cpp_daj_token_1(cshii, lok));
}

/* Maximum potrebnee na napiisanie tokenu (bez medzier). */
unsigned int cpp_token_dllzhka(const cpp_token *token)
{
    switch (piisanie_tokenov[token->typ].typ)
    {
        default: return (6);
        case PIISANIE_IDENT: return (token->hodnota.neviem->r.dllzhka * 10);
        case PIISANIE_LITERAAL: return (token->hodnota.rethazec.dllzhka);
    }
}

unsigned char *cpp_napiish_token(cpp_cshii *cshii, const cpp_token *token,
                                 unsigned char *pamaeth, int ktovie)
{
    switch (piisanie_tokenov[token->typ].typ)
    {
        case PIISANIE_INTERPUNKCIA: {
            const unsigned char *piisanie;
            unsigned char z;

            if (token->vlajky & DIGRAF)
            {
                piisanie = piisanie_digrafov[token->typ - CPP_PRVYY_DIGRAF];
            }
            else piisanie = piisanie_tokenov[token->typ].text;

            while ((z = *(piisanie++)) != '\0') *(pamaeth++) = z;

            break;
        }
        
        case PIISANIE_IDENT:
            if (1 /*ktovie +++DOROBITH univerzaalne znakovee menaa */)
            {
                memcpy(pamaeth,
                       token->hodnota.neviem->r.meno,
                       token->hodnota.neviem->r.dllzhka);
                pamaeth += token->hodnota.neviem->r.dllzhka;
            }

            break;
        
        case PIISANIE_LITERAAL:
            memcpy(pamaeth,
                   token->hodnota.rethazec.text,
                   token->hodnota.rethazec.dllzhka);
            pamaeth += token->hodnota.rethazec.dllzhka;
            break;
        
        case PIISANIE_ZHIADNE: printf("+++DOROB"); break;
    }

    return (pamaeth);
}

const unsigned char *cpp_token_ako_text(const cpp_token *token)
{
    switch (piisanie_tokenov[token->typ].typ)
    {
        case PIISANIE_INTERPUNKCIA:
            if (token->vlajky & DIGRAF)
            {
                return (piisanie_digrafov[token->typ - CPP_PRVYY_DIGRAF]);
            }
            else return (piisanie_tokenov[token->typ].text);
        case PIISANIE_IDENT: return (token->hodnota.neviem->r.meno);
        case PIISANIE_LITERAAL: return (token->hodnota.rethazec.text);
        case PIISANIE_ZHIADNE: return ((unsigned char *)"");
    }

    return ((unsigned char *)"Neznaamy typ!");
}

unsigned char *cpp_uuvodzovkuj_rethazec(unsigned char *cielh,
                                        const unsigned char *zdroj,
                                        unsigned int dllzhka)
{
    while (dllzhka--)
    {
        unsigned char z = *(zdroj++);

        switch (z)
        {
            case '\n':
                z = 'n';
                /* PREPAD. */

            case '\\':
            case '"':
                *(cielh++) = '\\';
        }

        *(cielh++) = z;
    }

    return (cielh);
}

unsigned char *cpp_daj_riadok_do_rethazca(cpp_cshii *cshii,
                                          const unsigned char *meno_direktiivy)
{
    const cpp_token *token;
    unsigned int von = 0;
    unsigned int alokovanee = 128;
    unsigned char *vyysledok = malloc(alokovanee);

    if (vyysledok == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    if (meno_direktiivy)
    {
        sprintf((char *)vyysledok, "#%s ", meno_direktiivy);
        von += strlen((const char *)meno_direktiivy) + 2;
    }

    token = cpp_daj_token(cshii);
    while (token->typ != CPP_KONIEC)
    {
        unsigned char *koniec;
        /* +2 pre medzeru a koncovyy '\0'. */
        unsigned int dllzhka = cpp_token_dllzhka(token) + 2;

        if (von + dllzhka > alokovanee)
        {
            alokovanee *= 2;
            if (von + dllzhka > alokovanee) alokovanee = von + dllzhka;

            vyysledok = realloc(vyysledok, alokovanee);
            if (vyysledok == NULL)
            {
                printf("Chyba pri alokaacii\n");
                abort();
            }
        }

        koniec = cpp_napiish_token(cshii, token, &(vyysledok[von]), 0);
        von = koniec - vyysledok;

        token = cpp_daj_token(cshii);
        if (token->vlajky & ZA_MEDZEROU) vyysledok[von++] = ' ';
    }

    vyysledok[von] = '\0';
    return (vyysledok); 
}

void cpp_vypiish_token(const cpp_token *token, FILE *vyystup)
{
    switch (piisanie_tokenov[token->typ].typ)
    {
        case PIISANIE_INTERPUNKCIA: {
            const unsigned char *piisanie;
            int z;

            if (token->vlajky & DIGRAF)
            {
                piisanie = piisanie_digrafov[token->typ - CPP_PRVYY_DIGRAF];
            }
            else piisanie = piisanie_tokenov[token->typ].text;
            
            z = *piisanie;
            do {
                putc(z, vyystup);
            } while ((z = *(++piisanie)) != '\0');

            break;
        }
        
        case PIISANIE_IDENT: {
            size_t i;
            const unsigned char *meno = NEVIEM_MENO(token->hodnota.neviem);

            for (i = 0; i < NEVIEM_DLLZHKA(token->hodnota.neviem); i++)
            {
                if (meno[i] & ~0x7F)
                {
                    printf("+++DOROB cpp_vypiish_token\n");
                }
                else putc(meno[i], vyystup);
            }

            break;
        }
        
        case PIISANIE_LITERAAL:
            fwrite(token->hodnota.rethazec.text, 1,
                   token->hodnota.rethazec.dllzhka, vyystup);
            break;
        
        case PIISANIE_ZHIADNE: break;
    }
}

int cpp_vyhni_sa_zlepeniu(cpp_cshii *cshii,
                          const cpp_token *token1,
                          const cpp_token *token2)
{
    enum cpp_typtokenu a = token1->typ, b = token2->typ;
    unsigned char z;

    z = ' ';
    if (token2->vlajky & DIGRAF)
    {
        z = piisanie_digrafov[b - CPP_PRVYY_DIGRAF][0];
    }
    else if (piisanie_tokenov[b].typ == PIISANIE_INTERPUNKCIA)
    {
        z = piisanie_tokenov[b].text[0];
    }

    /* Ryychlo naajde vshetko, csho sa mouzhe zlephith s '='. */
    if ((a <= CPP_P_POSUN) && (z == '=')) return (1);
    
    switch (a)
    {
        case CPP_VAECSHIE: return (z == '>');
        case CPP_MENSHIE: return ((z == '<') || (z == ':') || (z == '%'));
        case CPP_PLUS: return (z == '+');
        case CPP_MIINUS: return ((z == '-') || (z == '>'));
        case CPP_LOMKA: return ((z == '/') || (z == '*')); /* Komentaare. */
        case CPP_MODULO: return ((z == '>') || (z == ':'));
        case CPP_A: return (z == '&');
        case CPP_ALEBO: return (z == '|');
        case CPP_DVOJBODKA: return ((z == ':') || (z == '>'));
        case CPP_DEREF: return (z == '*');
        case CPP_BODKA: return ((z == '.') || (b == CPP_CSHIISLO));
        case CPP_MREZHA: return ((z == '#') || (z == '%')); /* Digraf. */
        case CPP_IDENT: return ((b == CPP_CSHIISLO)
                                || (b == CPP_IDENT)
                                || (b == CPP_ZNAK) || (CPP_RETHAZEC));
        case CPP_CSHIISLO: return ((b == CPP_CSHIISLO)
                                   || (b == CPP_IDENT)
                                   || (z == '.')
                                   || (z == '+')
                                   || (z == '-'));
        case CPP_INEE: return ((token1->hodnota.rethazec.text[0] == '\\')
                               && (b == CPP_IDENT));
    }
    
    return (0);
}

void *cpp_daj_pomoc(cpp_cshii *cshii)
{
    return (cshii->pomoc);
}

void *cpp_nastav_pomoc(cpp_cshii *cshii, void *pomoc)
{
    cshii->pomoc = pomoc;
}

void *cpp_daj_pomoc2(cpp_cshii *cshii)
{
    return (cshii->pomoc2);
}

void *cpp_nastav_pomoc2(cpp_cshii *cshii, void *pomoc)
{
    cshii->pomoc2 = pomoc;
}
