/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "cppkni.h"
#include "internee.h"
#include "podpora.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Akumulaator aktuaalne aktiivnych podmienok. */
struct ak_akumulaator {
    struct ak_akumulaator *dhalshii;
    int preskocsh_inak; /* Majuu sa buduuce #else a #elif preskocshith? */
    int preskakovalo_sa; /* preskakovanie == 1 pri pridanii? */
    int typ; /* Typ direktiivy. */
};  

typedef void (*spracovaavacsh_direktiivy)(cpp_cshii *);
typedef struct direktiiva {
    spracovaavacsh_direktiivy funkcia;
    const unsigned char *meno;
    unsigned int dllzhka;
    unsigned int vlajky;
} direktiiva;

/* Vlajky direktiiv. */
#define PODMIENKA 0x1 /* Podmienka. */
#define IF_PODMIENKA 0x2 /* Otvaaracia podmienka. */
#define INCLUDE 0x4 /* Cshiitajuu sa naazvy hlavicshiek. */
#define EXPAND 0x8 /* Vzhdy sa expandujuu tokeny. */

#define DIREKTIIVOVAA_TABULHKA \
D(if, D_IF, PODMIENKA | IF_PODMIENKA | EXPAND) \
D(ifdef, D_IFDEF, PODMIENKA | IF_PODMIENKA) \
D(ifndef, D_IFNDEF, PODMIENKA | IF_PODMIENKA) \
D(elif, D_ELIF, PODMIENKA | EXPAND) \
D(else, D_ELSE, PODMIENKA) \
D(endif, D_ENDIF, PODMIENKA) \
D(include, D_INCLUDE, INCLUDE | EXPAND) \
D(define, D_DEFINE, 0) \
D(undef, D_UNDEF, 0) \
D(line, D_LINE, EXPAND) \
D(error, D_ERROR, 0) \
D(pragma, D_PRAGMA, 0)

/* S pomocov tabulhky sa vygenerujuu prototypy, enum a shtruktuury. */
#define D(meno, emeno, vlajky) static void rob_##meno(cpp_cshii *);
DIREKTIIVOVAA_TABULHKA
#undef D
static void rob_oznacshovacsh_riadka(cpp_cshii *);

enum {
#define D(meno, emeno, vlajky) emeno,
    DIREKTIIVOVAA_TABULHKA
#undef D
    POCSHET_DIREKTIIV
};

#define D(meno, emeno, vlajky) {rob_##meno, (const unsigned char *)#meno, \
                                sizeof (#\
                                    meno) - 1, vlajky},
static const direktiiva direktiivy[] = {
    DIREKTIIVOVAA_TABULHKA
};
#undef D

static const direktiiva dir_oznacshovacsh_riadka = {
    rob_oznacshovacsh_riadka, (const unsigned char *)"#", 1, 0
};

/* Prototypy. */
static void zacshni_direktiivu(cpp_cshii *cshii);
static void ukoncshi_direktiivu(cpp_cshii *cshii, int preskocsh_riadok);
static void preskocsh_zvyshok_riadka(cpp_cshii *cshii);

/* +++NEPORIADOK. */
/* Skontroluje, cshi je makro v poriadku pre #define, #undef,
 * #ifdef a #ifndef. je_define_alebo_undef je 1, ak je toto volanee
 * z #define alebo #undef. */
static cpp_neviem *cshiitaj_meno_makra(cpp_cshii *cshii,
                                       int je_define_alebo_undef)
{
    cpp_token *token = _cpp_lex_token(cshii);

    if (token->typ == CPP_IDENT)
    {
        cpp_neviem *neviem = token->hodnota.neviem;

        if (je_define_alebo_undef
            && (neviem == cshii->shpec_neviem.n_defined))
        {
            /* C17 6.10.8.2. */
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "\"defined\" nemozhe byth pouzhitee ako naazov makra");
        }
        else return (neviem);
    }
    else if (token->typ == CPP_KONIEC)
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "zhiadne meno makra nebolo danee v direktiive #%s",
                  cshii->direktiiva->meno);
    }
    else
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "naazvy makier musia byth identifikaatory");
    }

    return (NULL);
}

/* Pomocnaa funkcia pre skontroluj_koniec_riadku. */
static void skontroluj_koniec_riadka_2(cpp_cshii *cshii, int expanduj)
{
    if ((cshii->akt_token[-1].typ != CPP_KONIEC)
        && ((expanduj
             ? cpp_daj_token(cshii)
             : _cpp_lex_token(cshii))
            ->typ != CPP_KONIEC))
    {
        cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                  "tokeny navyshe na konci direktiivy #%s",
                  cshii->direktiiva->meno);
    }
}

static void skontroluj_koniec_riadka_else_endif(cpp_cshii *cshii)
{
    skontroluj_koniec_riadka_2(cshii, 0);
}

/* Skontroluje, cshi suu na konci direktiivy tokeny navyshe.
 * Ak expanduj je 1, tokeny expandujuuce na nichs suu v poriadku. */
static void skontroluj_koniec_riadka(cpp_cshii *cshii, int expanduj)
{
    skontroluj_koniec_riadka_2(cshii, expanduj);
}

/* Daa token, ale preskocshii CPP_NICSH. */
static const cpp_token *daj_token_bez_nicsh(cpp_cshii *cshii)
{
    for (;;)
    {
        const cpp_token *vyysledok = cpp_daj_token(cshii);
        if (vyysledok->typ != CPP_NICSH) return (vyysledok);
    }
}

static unsigned char *zlep_naazov_hlavicshky(cpp_cshii *cshii)
{
    const cpp_token *token;
    unsigned char *pamaeth;
    size_t celkom = 0, kapacita = 2048 /* Naahodnee cshiislo. */;

    pamaeth = malloc(kapacita);
    if (pamaeth == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }
    for (;;)
    {
        size_t dllzhka;
        
        token = daj_token_bez_nicsh(cshii);
        if (token->typ == CPP_VAECSHIE) break;
        if (token->typ == CPP_KONIEC)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "chyyba ukoncshovacii znak >");
            break;
        }

        /* +2 pre medzeru pred a '\0' za. */
        dllzhka = cpp_token_dllzhka(token) + 2;
        if (celkom + dllzhka > kapacita)
        {
            kapacita = (kapacita + dllzhka) * 2;
            pamaeth = realloc(pamaeth, kapacita);
            if (pamaeth == NULL)
            {
                printf("Chyba pri alokaacii\n");
                abort();
            }
        }

        if (token->vlajky & ZA_MEDZEROU) pamaeth[celkom++] = ' ';

        celkom = (cpp_napiish_token(cshii, token, &(pamaeth[celkom]), 1)
                  - pamaeth);
    }

    pamaeth[celkom] = '\0';

    return (pamaeth);
}

static unsigned char *spracuj_include(cpp_cshii *cshii, int *shikmee)
{
    unsigned char *meno;
    const cpp_token *token;

    token = daj_token_bez_nicsh(cshii);
    if ((token->typ == CPP_RETHAZEC)
        || (token->typ == CPP_NAAZOV_HLAVICSHKY))
    {
        meno = malloc(token->hodnota.rethazec.dllzhka - 1);
        if (meno == NULL)
        {
            printf("Chyba pri alokaacii\n");
            abort();
        }
        memcpy(meno, token->hodnota.rethazec.text + 1,
               token->hodnota.rethazec.dllzhka - 2);
        meno[token->hodnota.rethazec.dllzhka - 2] = '\0';
        *shikmee = (token->typ == CPP_NAAZOV_HLAVICSHKY);
    }
    else if (token->typ == CPP_MENSHIE)
    {
        meno = zlep_naazov_hlavicshky(cshii);
        *shikmee = 1;
    }
    else
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "#%s ocshakaava \"NAAZOV SUUBORU\" alebo <NAAZOV SUUBORU>",
                  cshii->direktiiva->meno);
        return (NULL);
    }

    if (cshii->direktiiva == &(direktiivy[D_PRAGMA]))
        ;
    else skontroluj_koniec_riadka(cshii, 1);
    
    return (meno);
}

void spusth_direktiivu(cpp_cshii *cshii, int cshiislo_dir,
                       char *p, size_t dllzhka)
{
    _cpp_pridaj_ppos(cshii, (unsigned char *)p, dllzhka, 1 /* od_faazy3 */);
    zacshni_direktiivu(cshii);

    _cpp_cshisti_riadok(cshii);

    cshii->direktiiva = &(direktiivy[cshiislo_dir]);
    cshii->direktiiva->funkcia(cshii);

    ukoncshi_direktiivu(cshii, 1);

    _cpp_odstraanh_ppos(cshii);
}

/* Spracuje *reth, ako keby bol telom #define.
 * Ak je *reth len identifikaator, dostane hodnotu 1.
 * Inak maa byth formaat identifikaator=hodnota. */
void cpp_define(cpp_cshii *cshii, const char *reth)
{
    char *p;
    char *z;
    size_t dllzhka;
    
    dllzhka = strlen(reth);
    p = malloc(dllzhka + 3);
    if (p == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    memcpy(p, reth, dllzhka);
    z = strchr(reth, '=');
    if (z) p[z - reth] = ' ';
    else
    {
        p[dllzhka++] = ' ';
        p[dllzhka++] = '1';
    }
    p[dllzhka] = '\n';

    spusth_direktiivu(cshii, D_DEFINE, p, dllzhka);
    free(p);
}

void _cpp_define_vstavanee(cpp_cshii *cshii, const char *reth)
{
    char *p;
    size_t dllzhka = strlen(reth);
    p = malloc(dllzhka + 1);
    if (p == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    memcpy(p, reth, dllzhka);
    p[dllzhka] = '\n';

    spusth_direktiivu(cshii, D_DEFINE, p, dllzhka);
    free(p);
}

static void pridaj_podmienku(cpp_cshii *cshii, int preskocsh,
                             int typ, const cpp_neviem *neviem)
{
    struct ak_akumulaator *ak_aku;
    cpp_ppos *ppos = cshii->ppos;

    ak_aku = malloc(sizeof(*ak_aku));
    if (ak_aku == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    ak_aku->dhalshii = ppos->ak_akumulaator;
    ak_aku->preskocsh_inak = cshii->stav.preskakovanie || !preskocsh;
    ak_aku->preskakovalo_sa = cshii->stav.preskakovanie;
    ak_aku->typ = typ;

    cshii->stav.preskakovanie = preskocsh;
    ppos->ak_akumulaator = ak_aku;
}

static int rethazec_na_riadok(const unsigned char *text,
                              size_t dllzhka,
                              unsigned long *riadok,
                              int *pretecshenie)
{
    unsigned long r = 0, r_pred = 0;
    unsigned char z;

    *pretecshenie = 0;
    while (dllzhka--)
    {
        z = *(text++);
        if (!JECSHIISLICA(z)) return (1);

        r *= 10;
        r += z - '0';
        if (r < r_pred) *pretecshenie = 1;
        r_pred = r;
    }

    *riadok = r;

    return (0);
}    

/* Podfunkcia rob_oznacshovacsh_riadka(). Precshiita vlajky za naazvom suuboru.
 * predoshlaa je predoshlaa vlajka, 0, ak je to prvaa vlajka.
 * Vraati vlajku, ak je v poriadku. 0 pre koniec direktiivy. */
static unsigned int cshiitaj_vlajku(cpp_cshii *cshii, unsigned int predoshlaa)
{
    const cpp_token *token = _cpp_lex_token(cshii);

    if ((token->typ == CPP_CSHIISLO) && (token->hodnota.rethazec.dllzhka == 1))
    {
        unsigned int vlajka = token->hodnota.rethazec.text[0] - '0';

        if ((vlajka > predoshlaa) && (vlajka <= 4)
            && ((vlajka != 4) || (predoshlaa == 3))
            && ((vlajka != 2) || (predoshlaa == 0)))
        {
            return (vlajka);
        }
    }

    if (token->typ != CPP_KONIEC)
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "nespraavna vlajka \"%s\" v direktiive",
                  cpp_token_ako_text(token));
    }
    
    return (0);
}
/* +++KONIEC NEPORIADKU */

static void rob_if(cpp_cshii *cshii)
{
    int preskocsh = 1;

    if (!(cshii->stav.preskakovanie))
    {
        preskocsh = _cpp_spracuj_vyyraz(cshii, 1) == 0;
    }

    pridaj_podmienku(cshii, preskocsh, D_IF, 0);
}

static void rob_ifdef(cpp_cshii *cshii)
{
    int preskocsh = 1;

    if (!(cshii->stav.preskakovanie))
    {
        cpp_neviem *neviem = cshiitaj_meno_makra(cshii, 0);

        if (neviem)
        {
            preskocsh = !cpp_makro_s(neviem);
            skontroluj_koniec_riadka(cshii, 0);
        }
    }

    pridaj_podmienku(cshii, preskocsh, D_IFDEF, 0);
}

static void rob_ifndef(cpp_cshii *cshii)
{
    int preskocsh = 1;
    cpp_neviem *neviem = 0;

    if (!(cshii->stav.preskakovanie))
    {
        neviem = cshiitaj_meno_makra(cshii, 0);

        if (neviem)
        {
            preskocsh = cpp_makro_s(neviem);
            skontroluj_koniec_riadka(cshii, 0);
        }
    }

    pridaj_podmienku(cshii, preskocsh, D_IFNDEF, neviem);
}

static void rob_elif(cpp_cshii *cshii)
{
    struct ak_akumulaator *ak_aku = cshii->ppos->ak_akumulaator;

    if (ak_aku == NULL)
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "#elif bez #if");
    }
    else
    {
        if (ak_aku->typ == D_ELSE)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "#elif za #else");
        }
        ak_aku->typ = D_ELIF;

        if (ak_aku->preskocsh_inak)
            cshii->stav.preskakovanie = 1;
        else
        {
            cshii->stav.preskakovanie = _cpp_spracuj_vyyraz(cshii, 0) == 0;
            ak_aku->preskocsh_inak = !(cshii->stav.preskakovanie);
        }
    }
}

static void rob_else(cpp_cshii *cshii)
{
    struct ak_akumulaator *ak_aku = cshii->ppos->ak_akumulaator;

    if (ak_aku == NULL)
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "#else bez #if");
    }
    else
    {
        if (ak_aku->typ == D_ELSE)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "#else za #else");
        }
        ak_aku->typ = D_ELSE;

        /* Preskocshia sa vshetky buduuce (chybnee) #else a #elif. */
        cshii->stav.preskakovanie = ak_aku->preskocsh_inak;
        ak_aku->preskocsh_inak = 1;

        if (!(ak_aku->preskakovalo_sa))
            skontroluj_koniec_riadka_else_endif(cshii);
    }
}

static void rob_endif(cpp_cshii *cshii)
{
    cpp_ppos *ppos = cshii->ppos;
    struct ak_akumulaator *ak_aku = ppos->ak_akumulaator;

    if (ak_aku == NULL)
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "#endif bez #if");
    }
    else
    {
        if (!(ak_aku->preskakovalo_sa))
            skontroluj_koniec_riadka_else_endif(cshii);

        ppos->ak_akumulaator = ak_aku->dhalshii;
        cshii->stav.preskakovanie = ak_aku->preskakovalo_sa;
        free(ak_aku);
    }
}

static void rob_include(cpp_cshii *cshii)
{
    unsigned char *meno_suuboru;
    int shikmee; /* <meno> */

    meno_suuboru = spracuj_include(cshii, &shikmee);
    if (meno_suuboru == NULL) return;

    if (*meno_suuboru == '\0')
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "praazdnyy naazov suuboru v direktiive #%s",
                  cshii->direktiiva->meno);
        free(meno_suuboru);
        return;
    }

    /* Odstraani nahromadeeene neviem2. */
    preskocsh_zvyshok_riadka(cshii);

    _cpp_pridaj_include(cshii, (char *)meno_suuboru,
                        shikmee, TYP_INCLUDE_INCLUDE);

    free(meno_suuboru);
}

static void rob_define(cpp_cshii *cshii)
{
    cpp_neviem *neviem = cshiitaj_meno_makra(cshii, 1);

    if (neviem)
    {
        _cpp_definuj_makro(cshii, neviem);
    }
}

static void rob_undef(cpp_cshii *cshii)
{
    cpp_neviem *neviem = cshiitaj_meno_makra(cshii, 1);

    if (neviem)
    {
        if (cpp_makro_s(neviem))
        {
            _cpp_oddefinuj_makro(neviem);
        }
    }

    skontroluj_koniec_riadka(cshii, 0);
}

static void rob_line(cpp_cshii *cshii)
{
    const cpp_token *token;
    unsigned long riadok;
    unsigned long maximum = 2147483647;
    int pretecshenie;
    const char *novyy_suubor = cshii->ppos->meno_suuboru;

    token = cpp_daj_token(cshii);
    if ((token->typ != CPP_CSHIISLO)
        || rethazec_na_riadok(token->hodnota.rethazec.text,
                              token->hodnota.rethazec.dllzhka,
                              &riadok, &pretecshenie))
    {
        if (token->typ == CPP_KONIEC)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "necshakanyy koniec riadka za #line");
        }
        else
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "\"%s\" za #line nie je kladnee celee cshiislo",
                      cpp_token_ako_text(token));
        }
        return;
    }

    if ((riadok == 0) || (riadok > maximum) || (pretecshenie))
    {
        cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                  "cshiislo riadka mimo povolenyy rozsah");
    }

    token = cpp_daj_token(cshii);
    if (token->typ == CPP_RETHAZEC)
    {
        cpp_rethazec r = {0, NULL};
        if (cpp_interpretuj_rethazec(cshii, &(token->hodnota.rethazec),
                                     &r, CPP_RETHAZEC) == 0)
        {
            novyy_suubor = (const char *)(r.text);
        }
                                 
        skontroluj_koniec_riadka(cshii, /* expanduj */ 1);
    }
    else if (token->typ != CPP_KONIEC)
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "\"%s\" nie je spraavny naazov suuboru");
        return;
    }

    preskocsh_zvyshok_riadka(cshii);

    _cpp_zmenh_suubor(cshii, CPP_D_PREMENOVANIE, novyy_suubor, riadok);
}

static void rob_error(cpp_cshii *cshii)
{
    unsigned char *riadok;
    
    cshii->stav.blokuj_expanziu++;
    riadok = cpp_daj_riadok_do_rethazca(cshii, cshii->direktiiva->meno);
    cshii->stav.blokuj_expanziu--;

    cpp_chyba(cshii, CPP_DUU_CHYBA,
              "%s", riadok);
    free(riadok);
}

static void rob_pragma(cpp_cshii *cshii)
{
    printf("+++IMPLEMENTUJ rob_pragma\n");
}

static void rob_oznacshovacsh_riadka(cpp_cshii *cshii)
{
    const cpp_token *token;
    unsigned long riadok;
    int pretecshenie;
    const char *novyy_suubor = cshii->ppos->meno_suuboru;
    enum cpp_duovody duovod = CPP_D_PREMENOVANIE;
    unsigned int vlajka;

    /* Vraati token, nech sa daa znova precshiitath cshiislo. */
    _cpp_vraath_tokeny(cshii, 1);

    token = cpp_daj_token(cshii);
    if ((token->typ != CPP_CSHIISLO)
        || rethazec_na_riadok(token->hodnota.rethazec.text,
                              token->hodnota.rethazec.dllzhka,
                              &riadok, &pretecshenie))
    {
        /* Je nemozhnee tu dostat CPP_KONIEC,
         * takze cpp_token_ako_text() je bezpecshnee. */
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "\"%s\" za #line nie je kladnee celee cshiislo",
                  cpp_token_ako_text(token));
        return;
    }

    token = cpp_daj_token(cshii);
    if (token->typ == CPP_RETHAZEC)
    {
        cpp_rethazec r = {0, NULL};
        if (cpp_interpretuj_rethazec(cshii, &(token->hodnota.rethazec),
                                     &r, CPP_RETHAZEC) == 0)
        {
            novyy_suubor = (const char *)(r.text);
        }

        vlajka = cshiitaj_vlajku(cshii, 0);
        if (vlajka == 1)
        {
            duovod = CPP_D_VSTUP;
            /* Mohlo by sa tu niecsho urobith. */;
            vlajka = cshiitaj_vlajku(cshii, vlajka);
        }
        else if (vlajka == 2)
        {
            duovod = CPP_D_ODCHOD;
            vlajka = cshiitaj_vlajku(cshii, vlajka);
        }

        if (vlajka == 3)
        {
            /* +++DOROBITH */;
            vlajka = cshiitaj_vlajku(cshii, vlajka);
            if (vlajka == 4)
            {
                ;
            }
        }
                                 
        skontroluj_koniec_riadka(cshii, /* expanduj */ 1);
    }
    else if (token->typ != CPP_KONIEC)
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "\"%s\" nie je spraavny naazov suuboru");
        return;
    }

    preskocsh_zvyshok_riadka(cshii);

    _cpp_zmenh_suubor(cshii, duovod, novyy_suubor, riadok);
}

static void zacshni_direktiivu(cpp_cshii *cshii)
{
    cshii->stav.v_direktiive = 1;
    cshii->vyysledok_direktiivy.typ = CPP_NICSH;
}

/* Preskocshii ostaavajuuce tokeny v direktiive. */
static void preskocsh_zvyshok_riadka(cpp_cshii *cshii)
{
    /* Vyhodii nahromadenee neviem2. */
    while (cshii->neviem2 != NULL)
    {
        _cpp_odstraanh_neviem2(cshii);
    }

    if (cshii->akt_token[-1].typ != CPP_KONIEC)
    {
        while (_cpp_lex_token(cshii)->typ != CPP_KONIEC) ;
    }
}

static void ukoncshi_direktiivu(cpp_cshii *cshii, int preskocsh_riadok)
{
    if (preskocsh_riadok)
    {
        preskocsh_zvyshok_riadka(cshii);
        if (cshii->drzh_tokeny == 0)
        {
            cshii->akt_tokenrad = &(cshii->zaakladnyy_tokenrad);
            cshii->akt_token = cshii->akt_tokenrad->shtart;
        }
    }

    /* Obnovii stav. */
    cshii->stav.v_direktiive = 0;
    cshii->stav.zbieranie_naazvov_hlavicshiek = 0;
    cshii->direktiiva = 0;
}

/* Vraathi 0, ak direktiiva bola spracovanaa uplnee,
 * 1, ak sa maa riadok spracovath dhalej. */
int _cpp_spracuj_direktiivu(cpp_cshii *cshii)
{
    const direktiiva *dir = 0;
    const cpp_token *dmeno;
    int zbierali_sa_argumenty = cshii->stav.zbieranie_argumentov;
    int preskocsh = 1;

    if (zbierali_sa_argumenty)
    {
        cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                  "Nedefinovanee spraavanie,"
                  " direktiiva v argumentoch pre makro,"
                  " direktiiva sa spracuje");
        cshii->stav.zbieranie_argumentov = 0;
        cshii->stav.blokuj_expanziu = 0;
    }

    zacshni_direktiivu(cshii);
    dmeno = _cpp_lex_token(cshii);
    if (dmeno->typ == CPP_IDENT)
    {
        if (dmeno->hodnota.neviem->je_direktiiva)
            dir = &(direktiivy[dmeno->hodnota.neviem->index_direktiivy]);
    }
    else if (dmeno->typ == CPP_CSHIISLO)
    {
        dir = &dir_oznacshovacsh_riadka;
    }

    if (dir)
    {
        /* V preskocshenyych podmienkach suu vshetky nepodmienkovee direktiivy
         * ignorovanee, ale naazvy hlavicshiek
         * je potrebnee spraavne nacshiitath tak, cshi tak. */
        cshii->stav.zbieranie_naazvov_hlavicshiek = dir->vlajky & INCLUDE;
        cshii->stav.direktiiva_chce_medzery = dir->vlajky & INCLUDE;
        if (cshii->stav.preskakovanie && !(dir->vlajky & PODMIENKA))
        {
            dir = 0;
        }
    }
    else if (dmeno->typ == CPP_KONIEC) ; /* Nulovaa direktiiva. */
    else
    {
        if (cshii->stav.preskakovanie == 0)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "nespraavna preprocesorovaa direktiiva #%s",
                      cpp_token_ako_text(dmeno));
        }
    }

    cshii->direktiiva = dir;
    if (dir) cshii->direktiiva->funkcia(cshii);

    ukoncshi_direktiivu(cshii, preskocsh);
    if (zbierali_sa_argumenty)
    {
        /* Obnovii stav, ak sa zbierali argumenty pre makro. */
        cshii->stav.zbieranie_argumentov = 2;
        cshii->stav.blokuj_expanziu = 1;
    }

    return (!preskocsh);
}

/* Vlozhii vshetky direktiivy do cpp_neviem tabulhky. */
void _cpp_inic_direktiivy(cpp_cshii *cshii)
{
    unsigned int i;

    for (i = 0; i < POCSHET_DIREKTIIV; i++)
    {
        cpp_neviem *neviem = cpp_najdi(cshii, direktiivy[i].meno,
                                       direktiivy[i].dllzhka);

        neviem->je_direktiiva = 1;
        neviem->index_direktiivy = i;
    }
}

cpp_ppos *_cpp_pridaj_ppos(cpp_cshii *cshii,
                           unsigned char *shtart,
                           size_t velhkosth,
                           int od_faazy3)
{
    cpp_ppos *ppos = malloc(sizeof(*ppos));

    if (ppos == NULL)
    {
        printf("Zlyhanie pri alokaacii\n");
        abort();
    }

    memset(ppos, '\0', sizeof(*ppos));

    ppos->dhalshii_riadok = ppos->shtart = shtart;
    ppos->koniec = shtart + velhkosth;
    ppos->od_faazy3 = od_faazy3;
    ppos->potrebnyy_dhalshii_riadok = 1;
    ppos->predoshlyy = cshii->ppos;

    ppos->suubor = NULL;

    cshii->ppos = ppos;

    return (ppos);
}

void _cpp_odstraanh_ppos(cpp_cshii *cshii)
{
    cpp_ppos *staryy = cshii->ppos;
    struct ak_akumulaator *ak_aku, *dhalshii_ak_aku;

    for (ak_aku = staryy->ak_akumulaator;
         ak_aku;
         ak_aku = dhalshii_ak_aku)
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "neukoncshenee #%s",
                  direktiivy[ak_aku->typ].meno);
        dhalshii_ak_aku = ak_aku->dhalshii;
        free(ak_aku);
    }
    
    /* Ak by chyybal #endif. */
    cshii->stav.preskakovanie = 0;

    cshii->ppos = staryy->predoshlyy;

    if (cshii->ppos)
    {
        if (staryy->suubor)
        {
            _cpp_zmenh_suubor(cshii, CPP_D_ODCHOD,
                              cshii->ppos->meno_suuboru,
                              cshii->ppos->riadok);
        }
    }

    free(staryy->poznaamky);
    free((void *)(staryy->shtart));
    free(staryy);
}

void _cpp_zmenh_suubor(cpp_cshii *cshii, enum cpp_duovody duovod,
                       const char *suubor, unsigned long riadok)
{
    lokaacia_t lok;
    
    lok.riadok = riadok;
    lok.suubor = suubor;

    cshii->ppos->riadok = riadok;
    cshii->ppos->meno_suuboru = suubor;
            
    if (cshii->spaetneevolania.zmena_suuboru)
    {
        cshii->spaetneevolania.zmena_suuboru(cshii, lok, duovod);
    }
}
