/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "cppkni.h"
#include "internee.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

unsigned char _cpp_trigrafovaa_mapa[0x100];

static void inic_trigrafovuu_mapu(void)
{
    _cpp_trigrafovaa_mapa['='] = '#';
    _cpp_trigrafovaa_mapa[')'] = ']';
    _cpp_trigrafovaa_mapa['!'] = '|';
    _cpp_trigrafovaa_mapa['('] = '[';
    _cpp_trigrafovaa_mapa['\''] = '^';
    _cpp_trigrafovaa_mapa['>'] = '}';
    _cpp_trigrafovaa_mapa['/'] = '\\';
    _cpp_trigrafovaa_mapa['<'] = '{';
    _cpp_trigrafovaa_mapa['-'] = '~';
}

static void inic(void)
{
    static int inicializovanee = 0;

    if (inicializovanee == 0)
    {
        inicializovanee = 1;

        inic_trigrafovuu_mapu();
    }
}

void cpp_nastav_jazyk(cpp_cshii *cshii, enum jazyk_c jazyk)
{
    CPP_NASTAVENIE(cshii, jazyk) = jazyk;
    
    CPP_NASTAVENIE(cshii, trigrafy) = 1;
    CPP_NASTAVENIE(cshii, rozshiirenee_cshiisla) = (jazyk == SHTD_C17) ? 1 : 0;
}

cpp_cshii *cpp_vytvor_cshii(enum jazyk_c jazyk, tabsym *tabulhka)
{
    cpp_cshii *cshii;

    inic();

    cshii = malloc(sizeof(*cshii));
    if (cshii == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }
    memset(cshii, '\0', sizeof(*cshii));

    cpp_nastav_jazyk(cshii, jazyk);
    CPP_NASTAVENIE(cshii, presnosth) = CHAR_BIT * sizeof(long);

    /* Zaakladne sa lokaacia nevynucuje. */
    cshii->nuutenaa_lokaacia.suubor = NULL;

    /* Nastavii statickee tokeny. */
    cshii->koniec.typ = CPP_KONIEC;
    cshii->koniec.vlajky = 0;
    cshii->blokuj_zlepenie.typ = CPP_NICSH;
    cshii->blokuj_zlepenie.hodnota.zdroj = NULL;

    /* Pripravii pamaeth pre tokeny. */
    _cpp_inic_tokenrad(&(cshii->zaakladnyy_tokenrad));
    cshii->akt_tokenrad = &(cshii->zaakladnyy_tokenrad);
    cshii->akt_token = cshii->akt_tokenrad->shtart;

    /* Faloshnyy praazdny adresaar pre suubory bez hlhadacej cesty. */
    cshii->zhiadna_hlhadacia_cesta.meno = "";
    cshii->zhiadna_hlhadacia_cesta.dhalshii = NULL;

    cshii->uuvodzovkovee_include = NULL;
    cshii->shikmee_include = NULL;

    cshii->uuvodzovky_ignorujuu_zdrojovyy_adresaar = 0;

    /* Akumulaator pre spracovanie vyyrazov. */
    cshii->op_akumulaator = NULL;
    cshii->op_koniec = NULL;
    _cpp_expanduj_op_akumulaator(cshii);

    _cpp_inic_tabsym(cshii, tabulhka);

    cshii->daatum = NULL;
    cshii->cshas = NULL;

    return (cshii);
}

void cpp_znicsh_cshii(cpp_cshii *cshii)
{
    if (cshii->vlastnaa_tabulhka) tabsym_znicsh_tabsym(cshii->tab);
    free(cshii->daatum);
    free(cshii->cshas);
    free(cshii);
}

struct vstavanee_makro {
    const unsigned char *meno;
    const unsigned int dllzhka;
    const unsigned int hodnota;
};

#define RETHAZEC_A_DLLZHKA(r) (const unsigned char *)r, sizeof(r) - 1
#define A(meno, hodnota) {RETHAZEC_A_DLLZHKA(meno), hodnota}

static const struct vstavanee_makro vstavanee[] = {
    A("__DATE__", VS_DATE),
    A("__FILE__", VS_FILE),
    A("__LINE__", VS_LINE),
    A("__TIME__", VS_TIME),
    A("_Pragma", VS_PRAGMA)
};

void cpp_inic_shpec_vstavanee(cpp_cshii *cshii)
{
    const struct vstavanee_makro *vs;
    size_t d = sizeof(vstavanee) / sizeof(*vstavanee);

    for (vs = vstavanee; vs < vstavanee + d; vs++)
    {
        cpp_neviem *neviem = cpp_najdi(cshii, vs->meno, vs->dllzhka);
        neviem->typ = NEVIEM_VSTAVANEE_MAKRO;
        neviem->hodnota.vstavanee = vs->hodnota;
    }
}

void cpp_inic_vstavanee(cpp_cshii *cshii, int hosthovanee)
{
    cpp_inic_shpec_vstavanee(cshii);
    
    _cpp_define_vstavanee(cshii, "__STDC__ 1");

    if (CPP_NASTAVENIE(cshii, jazyk) == SHTD_C17)
    {
        _cpp_define_vstavanee(cshii, "__STDC_VERSION__ 201710L");
    }

    if (hosthovanee) _cpp_define_vstavanee(cshii, "__STDC_HOSTED__ 1");
    else _cpp_define_vstavanee(cshii, "__STDC_HOSTED__ 0");
}

const char *cpp_cshiitaj_hlavnyy_suubor(cpp_cshii *cshii, const char *meno)
{
    cshii->hlavnyy_suubor = _cpp_najdi_suubor(cshii,
                                              meno,
                                              &(cshii->zhiadna_hlhadacia_cesta)
                                              );
    if (_cpp_suubor_nenajdenyy(cshii->hlavnyy_suubor)) return (NULL);
    if (_cpp_pridaj_suubor(cshii, cshii->hlavnyy_suubor)) return (NULL);
    
    return (meno);
}
