/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "cppkni.h"
#include "internee.h"
#include "podpora.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define PRESNOSTH_CSHASTI (sizeof(cpp_cshasth_cshiisla) * CHAR_BIT)
#define MASKA_POLOVICE ((~(cpp_cshasth_cshiisla)0) >> (PRESNOSTH_CSHASTI / 2))
#define DOLNAA_CSHASTH(a) ((a) & MASKA_POLOVICE)
#define HORNAA_CSHASTH(a) ((a) >> (PRESNOSTH_CSHASTI / 2))

struct op {
    const cpp_token *token;
    cpp_cshiislo hodnota; /* Hodnota napravo od operaatora. */
    enum cpp_typtokenu op;
};

#define CPP_UPLUS ((enum cpp_typtokenu) CPP_POSLEDNYY_CPP_OP + 1)
#define CPP_UMIINUS ((enum cpp_typtokenu) CPP_POSLEDNYY_CPP_OP + 2)

#define cshiislo_je_nula(a) (((a).hornaa == 0) && ((a).dolnaa == 0))
#define cshiisla_suu_rovnee(a, b) (((a).hornaa == (b).hornaa) \
                                   && ((a).dolnaa == (b).dolnaa))

static unsigned int interpretuj_desatinnuu_priiponu(cpp_cshii *cshii,
                                                    const unsigned char *reth,
                                                    size_t dllzhka)
{
    size_t f, l;

    f = l = 0;

    while (dllzhka--)
    {
        switch(reth[0])
        {
            case 'f':
            case 'F':
                f++;
                break;

            case 'l':
            case 'L':
                l++;
                break;

            default: return (0);
        }
        reth++;
    }

    if (f + l > 1) return (0);
    
    return (f ? CPP_CSHIISLO_MALEE :
            l ? CPP_CSHIISLO_VELHKEE :
            CPP_CSHIISLO_ZAAKLADNEE);
}

static unsigned int interpretuj_celuu_priiponu(cpp_cshii *cshii,
                                               const unsigned char *reth,
                                               size_t dllzhka)
{
    size_t l, u;

    l = u = 0;

    while (dllzhka--)
    {
        switch(reth[dllzhka])
        {
            case 'u':
            case 'U':
                u++;
                break;
            
            case 'l':
            case 'L':
                l++;
                /* Ak suu 2 l-kaa, musia byth za sebou a rovnakej velhkosti. */
                if ((l == 2) && (reth[dllzhka] != reth[dllzhka + 1])) return (0);
                break;

            default: return (0);
        }
    }

    if ((l > 2) || (u > 1)) return (0);
    
    return ((u ? CPP_CSHIISLO_BEZZNAMIENKOVEE : 0)
            | ((l == 0) ? CPP_CSHIISLO_MALEE :
               (l == 1) ? CPP_CSHIISLO_STREDNEE :
               CPP_CSHIISLO_VELHKEE));
}

static unsigned int hex_hodnota(unsigned int z)
{
    if ((z >= '0') && (z <= '9')) return (z - '0');
    if ((z >= 'a') && (z <= 'f')) return (z + 10 - 'a');
    if ((z >= 'A') && (z <= 'F')) return (z + 10 - 'A');

    printf("%c nie je v hexadecimaalnej suustave\n", z);
    return (17);
}

unsigned int cpp_klasifikuj_cshiislo(cpp_cshii *cshii, const cpp_token *token)
{
    const unsigned char *reth = token->hodnota.rethazec.text;
    const unsigned char *koniec;
    unsigned int vyysledok, korenh, najvaecshia_cshiislica;
    enum {NIE_DESATINNEE, ZA_BODKOU, ZA_EXPONENTOM} desatinnaa_vlajka;
    int videnaa_cshiislica;

    /* Ak je dllzhka 1, cshiislo je len jedna cifra. */
    if (token->hodnota.rethazec.dllzhka == 1)
        return (CPP_CSHIISLO_CELEE | CPP_CSHIISLO_MALEE | CPP_CSHIISLO_DESIAT);

    koniec = reth + (token->hodnota.rethazec.dllzhka);
    desatinnaa_vlajka = NIE_DESATINNEE;
    korenh = 10;
    najvaecshia_cshiislica = 0;
    videnaa_cshiislica = 0;

    /* Najprv sa zistii korenh (suustava). */
    if (*reth == '0')
    {
        korenh = 8;
        reth++;

        switch (*reth)
        {
            case 'x':
            case 'X':
                if ((reth[1] == '.') || JEXCSHIISLICA(reth[1]))
                {
                    korenh = 16;
                    reth++;
                }
                break;
            
            case 'b':
            case 'B':
                if ((reth[1] == '0') || (reth[1] == '1'))
                {
                    korenh = 2;
                    reth++;
                }
                break;
        }
    }

    for (;;)
    {
        unsigned int z = *(reth++);

        if (JECSHIISLICA(z) || (JEXCSHIISLICA(z) && (korenh == 16)))
        {
            videnaa_cshiislica = 1;
            z = hex_hodnota(z);
            if (z > najvaecshia_cshiislica) najvaecshia_cshiislica = z;
        }
        else if (z == '.')
        {
            if (desatinnaa_vlajka == NIE_DESATINNEE)
                desatinnaa_vlajka = ZA_BODKOU;
            else
            {
                cpp_chyba(cshii, CPP_DUU_CHYBA,
                          "priilish velha desatinnnych bodiek v cshiisle");
                return (CPP_CSHIISLO_NESPRAAVNE);
            }
        }
        else if (((korenh <= 10) && ((z == 'e') || (z == 'E')))
                 || ((korenh == 16) && ((z == 'p') || (z == 'P'))))
        {
            desatinnaa_vlajka = ZA_EXPONENTOM;
            break;
        }
        else
        {
            /* Zacshiatok priipony. */
            reth--;
            break;
        }
    }

    if ((desatinnaa_vlajka != NIE_DESATINNEE) && (korenh == 8)) korenh = 10;

    /* Muozhe nastath len pre binaarne a oktaalne konshtanty. */
    if (najvaecshia_cshiislica > korenh)
    {
        if (korenh == 2)
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "nespraavna cshiislica \"%c\" v binaarnej konshtante",
                      '0' + najvaecshia_cshiislica);
        else
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "nespraavna cshiislica \"%c\" v oktaalnej konshtante",
                      '0' + najvaecshia_cshiislica);
        return (CPP_CSHIISLO_NESPRAAVNE);
    }

    if (desatinnaa_vlajka != NIE_DESATINNEE)
    {
        if (korenh == 2)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "nespraavna predpona \"0b\" pre desatinnne cshiislo");
            return (CPP_CSHIISLO_NESPRAAVNE);
        }

        if ((korenh == 16) && !videnaa_cshiislica)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "zhiadne cshiislice v hexadecimaalnej"
                      " desatinnej konsthante");
            return (CPP_CSHIISLO_NESPRAAVNE);
        }

        if ((korenh == 16) && !CPP_NASTAVENIE(cshii, rozshiirenee_cshiisla))
        {
            cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                      "pouzhitie C17 hexadecimaalnej desatinnej konsthanty");
        }

        if (desatinnaa_vlajka == ZA_EXPONENTOM)
        {
            if ((*reth == '+') || (*reth == '-')) reth++;

            /* Exponent je vzhdy v desiatkovej suustave. */
            if (!JECSHIISLICA(*reth))
            {
                cpp_chyba(cshii, CPP_DUU_CHYBA,
                          "exponent nemaa zhiadnee cshiislice");
                return (CPP_CSHIISLO_NESPRAAVNE);
            }

            do {
                reth++;
            } while (JECSHIISLICA(*reth));
        }
        else if (korenh == 16)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "hexadecimaalne desatinnee konshtanty musia math exponent");
            return (CPP_CSHIISLO_NESPRAAVNE);
        }

        vyysledok = interpretuj_desatinnuu_priiponu(cshii, reth, koniec - reth);
        if (vyysledok == 0)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "nespraavna priipona \"%s\" na desatinnej konshtante",
                      reth);
            return (CPP_CSHIISLO_NESPRAAVNE);
        }

        vyysledok |= CPP_CSHIISLO_DESATINNEE;
    }
    else
    {
        vyysledok = interpretuj_celuu_priiponu(cshii, reth, koniec - reth);
        if (vyysledok == 0)
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "nespraavna priipona \"%s\" na celej konshtante",
                      reth);
            return (CPP_CSHIISLO_NESPRAAVNE);
        }

        vyysledok |= CPP_CSHIISLO_CELEE;
    }

    switch (korenh)
    {
        case 10: vyysledok |= CPP_CSHIISLO_DESIAT; break;
        case 16: vyysledok |= CPP_CSHIISLO_HEX; break;
        case 8: vyysledok |= CPP_CSHIISLO_OKTAALNE; break;
        case 2: vyysledok |= CPP_CSHIISLO_BINAARNE; break;
    }

    return (vyysledok);
}

static cpp_cshiislo cshiislo_orezh(cpp_cshiislo vyysledok, size_t presnosth)
{
    if (presnosth > PRESNOSTH_CSHASTI)
    {
        presnosth -= PRESNOSTH_CSHASTI;
        if (presnosth < PRESNOSTH_CSHASTI)
            vyysledok.hornaa &= (((cpp_cshasth_cshiisla)1) << presnosth) - 1;
    }
    else
    {
        if (presnosth < PRESNOSTH_CSHASTI)
            vyysledok.dolnaa &= (((cpp_cshasth_cshiisla)1) << presnosth) - 1;
        vyysledok.hornaa = 0;
    }
    
    return (vyysledok);
}

static cpp_cshiislo pridaj_cshiislicu(cpp_cshiislo csh, unsigned int z,
                                      unsigned int zaaklad, size_t presnosth)
{
    cpp_cshiislo vyysledok;
    unsigned int posun;
    int pretecshenie;
    cpp_cshasth_cshiisla pridanie_hornee, pridanie_dolnee;

    switch (zaaklad)
    {
        case 2: posun = 1; break;
        case 16: posun = 4; break;
        default: posun = 3; break;
    }

    vyysledok.pretecshenie = !!(csh.hornaa >> (PRESNOSTH_CSHASTI - posun));
    vyysledok.hornaa = csh.hornaa << posun;
    vyysledok.dolnaa = csh.dolnaa << posun;
    vyysledok.hornaa |= csh.dolnaa >> (PRESNOSTH_CSHASTI - posun);
    vyysledok.bezznamienkovee = csh.bezznamienkovee;

    if (zaaklad == 10)
    {
        pridanie_dolnee = csh.dolnaa << 1;
        pridanie_hornee = ((csh.hornaa << 1)
                           + (csh.dolnaa >> (PRESNOSTH_CSHASTI - 1)));
    }
    else pridanie_dolnee = pridanie_hornee = 0;

    if (pridanie_dolnee + z < pridanie_dolnee) pridanie_hornee++;
    pridanie_dolnee += z;

    if (vyysledok.dolnaa + pridanie_dolnee < vyysledok.dolnaa)
        pridanie_hornee++;
    if (vyysledok.hornaa + pridanie_hornee < vyysledok.hornaa)
        vyysledok.pretecshenie = 1;

    vyysledok.dolnaa += pridanie_dolnee;
    vyysledok.hornaa += pridanie_hornee;

    /* */
    csh.dolnaa = vyysledok.dolnaa;
    csh.hornaa = vyysledok.hornaa;
    vyysledok = cshiislo_orezh(vyysledok, presnosth);
    if (!cshiisla_suu_rovnee(vyysledok, csh))
        vyysledok.pretecshenie = 1;
    
    return (vyysledok);
}

static int cshiislo_kladnee(cpp_cshiislo csh, size_t presnosth)
{
    if (presnosth > PRESNOSTH_CSHASTI)
    {
        presnosth -= PRESNOSTH_CSHASTI;
        return ((csh.hornaa
                 & (((cpp_cshasth_cshiisla)1) << (presnosth - 1))) == 0);
    }
    return ((csh.dolnaa
             & (((cpp_cshasth_cshiisla)1) << (presnosth - 1))) == 0);
}

cpp_cshiislo cpp_interpretuj_celee_cshiislo(cpp_cshii *cshii,
                                            const cpp_token *token,
                                            unsigned int typ)
{
    cpp_cshiislo vyysledok;
    const unsigned char *p;

    vyysledok.bezznamienkovee = !!(typ & CPP_CSHIISLO_BEZZNAMIENKOVEE);
    vyysledok.pretecshenie = 0;
    vyysledok.hornaa = vyysledok.dolnaa = 0;

    p = token->hodnota.rethazec.text;
    
    if (token->hodnota.rethazec.dllzhka == 1)
    {
        vyysledok.dolnaa = p[0] - '0';
    }
    else
    {
        cpp_cshasth_cshiisla max;
        size_t presnosth = CPP_NASTAVENIE(cshii, presnosth);
        const unsigned char *koniec = p + (token->hodnota.rethazec.dllzhka);
        unsigned int zaaklad = 10;
        int pretecshenie = 0;

        switch (typ & CPP_CSHIISLO_KORENH)
        {
            case CPP_CSHIISLO_OKTAALNE: zaaklad = 8; p++; break;
            case CPP_CSHIISLO_HEX: zaaklad = 16; p += 2; break;
            case CPP_CSHIISLO_BINAARNE: zaaklad = 2; p += 2; break;
        }

        max = ~(cpp_cshasth_cshiisla)0;
        if (presnosth < PRESNOSTH_CSHASTI)
        {
            max >>= PRESNOSTH_CSHASTI - presnosth;
        }
        max = (max - zaaklad + 1) / zaaklad + 1;

        for (; p < koniec; p++)
        {
            unsigned int z = *p;

            if (JECSHIISLICA(z) || (JEXCSHIISLICA(z) && (zaaklad == 16)))
            {
                z = hex_hodnota(z);
            }
            else break;

            if (vyysledok.dolnaa < max)
            {
                vyysledok.dolnaa = vyysledok.dolnaa * zaaklad + z;
            }
            else
            {
                vyysledok = pridaj_cshiislicu(vyysledok, z, zaaklad, presnosth);
                pretecshenie |= vyysledok.pretecshenie;
                max = 0;
            }
        }

        if (pretecshenie)
        {
            cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                      "celocshiiselnaa konshtanta"
                      " je priliish velhkaa na svoj typ");
        }
        else if (!(vyysledok.bezznamienkovee)
                 && !cshiislo_kladnee(vyysledok, presnosth))
        {
            if (zaaklad == 10)
                cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                          "celocshiiselnaa konshtanta je taaka velhkaa,"
                          " zhe je bezznamiekovaa");
            vyysledok.bezznamienkovee = 1;
        }
    }
    
    return (vyysledok);
}

cpp_cshiislo spracuj_defined(cpp_cshii *cshii)
{
    cpp_cshiislo vyysledok;
    const cpp_token *token;
    int zaatvorka = 0;
    cpp_neviem *neviem = NULL;
    cpp_neviem2 *pocshiatocshnee_neviem2 = cshii->neviem2;

    cshii->stav.blokuj_expanziu++;

    token = cpp_daj_token(cshii);
    if (token->typ == CPP_LH_ZAATVORKA)
    {
        zaatvorka = 1;
        token = cpp_daj_token(cshii);
    }

    if (token->typ == CPP_IDENT)
    {
        neviem = token->hodnota.neviem;
        if (zaatvorka && (cpp_daj_token(cshii)->typ != CPP_P_ZAATVORKA))
        {
            cpp_chyba(cshii, CPP_DUU_CHYBA,
                      "chyyba ')' za \"defined\"");
            neviem = NULL;
        }
    }
    else
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "operaator \"defined\" potrebuje identifikaator");
    }

    if (neviem)
    {
        if ((cshii->neviem2 != pocshiatocshnee_neviem2)
            || (pocshiatocshnee_neviem2 != NULL))
        {
            cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                      "toto pouzhitie \"defined\" nemusii vshade fungovath"
                      " (expanzia na \"defined\")");
        }
    }

    cshii->stav.blokuj_expanziu--;

    vyysledok.bezznamienkovee = 0;
    vyysledok.pretecshenie = 0;
    vyysledok.hornaa = 0;
    vyysledok.dolnaa = (neviem && cpp_makro_s(neviem));
    
    return (vyysledok);
}

static cpp_cshiislo vyhodnoth_token(cpp_cshii *cshii, const cpp_token *token)
{
    cpp_cshiislo vyysledok;

    vyysledok.bezznamienkovee = 0;
    vyysledok.pretecshenie = 0;
    
    switch (token->typ)
    {
        case CPP_CSHIISLO: {
            unsigned int typ = cpp_klasifikuj_cshiislo(cshii, token);

            switch (typ & CPP_CSHIISLO_KATEGOORIA)
            {
                case CPP_CSHIISLO_DESATINNEE:
                    /* C17 6.10.1.1 */
                    cpp_chyba(cshii, CPP_DUU_CHYBA,
                              "desatinnaa konshtanta v preprocesorovm vyyraze");
                    break;

                case CPP_CSHIISLO_CELEE:
                    return (cpp_interpretuj_celee_cshiislo(cshii, token, typ));

                case CPP_CSHIISLO_NESPRAAVNE:
                    break;
            }

            vyysledok.hornaa = vyysledok.dolnaa = 0;
            break;
        }

        case CPP_ZNAK:
        case CPP_LZNAK:
        case CPP_uZNAK:
        case CPP_UZNAK: {
            printf("+++Znakovee konshtanty nie suu vo vyyrazoch zatialh"
                   " podporovanee\n");
            vyysledok.hornaa = vyysledok.dolnaa = 0;
            break;
        }
        
        case CPP_IDENT:
            if (token->hodnota.neviem == (cshii->shpec_neviem.n_defined))
                return (spracuj_defined(cshii));
            vyysledok.hornaa = vyysledok.dolnaa = 0;
            break;
    }

    return (vyysledok);
}

#define ZHIADNY_LH_OPERAND 0x1
#define LHAVO_ASOCIATIIVNE 0x2
#define SKONTROLUJ_POVYYSHENIE 0x4

static const struct cpp_operaator {
    unsigned int priorita;
    unsigned int vlajky;
} optab[] = {
    /* ROVNAA_SA */ {0, 0}, /* Nemaa sa stath. */
    /* NIE */ {16, ZHIADNY_LH_OPERAND},
    /* VAECSHIE */ {12, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* MENSHIE */ {12, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* PLUS */ {13+1, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* MIINUS */ {13+1, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* HVIEZDA */ {15, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* LOMKA */ {15, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* MODULO */ {15, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* A */ {9, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* ALEBO */ {7, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* XALEBO */ {8, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* LH_POSUN */ {13, LHAVO_ASOCIATIIVNE},
    /* P_POSUN */ {13, LHAVO_ASOCIATIIVNE},

    /* KOMPLEMENT */ {16, ZHIADNY_LH_OPERAND},
    /* A_A */ {6, LHAVO_ASOCIATIIVNE},
    /* ALEBO_ALEBO */ {5, LHAVO_ASOCIATIIVNE},
    /* OTAAZNIK */ {4, 0},
    /* DVOJBODKA */ {4, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* CSHIARKA */ {4, LHAVO_ASOCIATIIVNE},
    /* LH_ZAATVORKA */ {1, ZHIADNY_LH_OPERAND},
    /* P_ZAATVORKA */ {0, 0},
    /* KONIEC */ {0, 0},
    /* ROVNAA_SA_ROVNAA_SA */ {11, LHAVO_ASOCIATIIVNE},
    /* NIE_ROVNAA_SA */ {11, LHAVO_ASOCIATIIVNE},
    /* VAECSHIE_ROVNAA_SA */ {12, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* MENSHIE_ROVNAA_SA */ {12, LHAVO_ASOCIATIIVNE | SKONTROLUJ_POVYYSHENIE},
    /* UPLUS */ {16, ZHIADNY_LH_OPERAND},
    /* UMIINUS */ {16, ZHIADNY_LH_OPERAND}
};

static skontroluj_povyyshenie(cpp_cshii *cshii, struct op *op)
{
    if (op->hodnota.bezznamienkovee == (op[-1].hodnota.bezznamienkovee))
        return;

    if (op->hodnota.bezznamienkovee)
    {
        if (!cshiislo_kladnee(op[-1].hodnota, CPP_NASTAVENIE(cshii, presnosth)))
        {
            printf("lhavyy operand \"%s\" menii znamienko pri povyyshenii\n",
                   cpp_token_ako_text(op->token));
        }
    }
    else if (!cshiislo_kladnee(op->hodnota, CPP_NASTAVENIE(cshii, presnosth)))
    {
        printf("pravyy operand \"%s\" menii znamienko pri povyyshenii\n",
               cpp_token_ako_text(op->token));
    }
}

static cpp_cshiislo cshiislo_zneguj(cpp_cshiislo csh, size_t presnosth)
{
    cpp_cshiislo koopia = csh;

    csh.dolnaa = ~(csh.dolnaa);
    csh.hornaa = ~(csh.hornaa);
    if (++(csh.dolnaa) == 0)
        csh.hornaa++;
    csh = cshiislo_orezh(csh, presnosth);
    csh.pretecshenie = (!(csh.bezznamienkovee)
                        && cshiisla_suu_rovnee(csh, koopia)
                        && !cshiislo_je_nula(csh));

    return (csh);
}

static cpp_cshiislo cshiislo_unaarna_op(cpp_cshii *cshii, cpp_cshiislo csh,
                                        enum cpp_typtokenu op)
{
    switch (op)
    {
        case CPP_UPLUS:
            csh.pretecshenie = 0;
            break;

        case CPP_UMIINUS:
            csh = cshiislo_zneguj(csh, CPP_NASTAVENIE(cshii, presnosth));
            break;

        case CPP_NIE:
            csh.dolnaa = cshiislo_je_nula(csh);
            csh.hornaa = 0;
            csh.bezznamienkovee = 0;
            csh.pretecshenie = 0;
            break;

        case CPP_KOMPLEMENT:
            csh.dolnaa = ~(csh.dolnaa);
            csh.hornaa = ~(csh.hornaa);
            csh = cshiislo_orezh(csh, CPP_NASTAVENIE(cshii, presnosth));
            csh.pretecshenie = 0;
            break;

        default:
            printf("Nespraavnaa operaacia v cshiislo_unaarna_op\n");
            break;
    }

    return (csh);
}

static cpp_cshiislo cshiislo_p_posun(cpp_cshiislo csh,
                                     size_t presnosth,
                                     size_t posun)
{
    cpp_cshasth_cshiisla znamienkovee_n;

    if (csh.bezznamienkovee || cshiislo_kladnee(csh, presnosth))
        znamienkovee_n = 0;
    else znamienkovee_n = ~(cpp_cshasth_cshiisla)0;

    if (posun >= presnosth)
        csh.hornaa = csh.dolnaa = znamienkovee_n;
    else
    {
        if (presnosth < PRESNOSTH_CSHASTI)
        {
            csh.hornaa = znamienkovee_n;
            csh.dolnaa |= znamienkovee_n << presnosth;
        }
        else if (presnosth < 2 * PRESNOSTH_CSHASTI)
            csh.hornaa |= znamienkovee_n << (presnosth - PRESNOSTH_CSHASTI);

        if (posun >= PRESNOSTH_CSHASTI)
        {
            posun -= PRESNOSTH_CSHASTI;
            csh.dolnaa = csh.hornaa;
            csh.hornaa = znamienkovee_n;
        }

        if (posun)
        {
            csh.dolnaa = ((csh.dolnaa >> posun)
                          | (csh.hornaa << (PRESNOSTH_CSHASTI - posun)));
            csh.hornaa = ((csh.hornaa >> posun)
                          | (znamienkovee_n << (PRESNOSTH_CSHASTI - posun)));
        }
    }

    csh = cshiislo_orezh(csh, presnosth);
    csh.pretecshenie = 0;
    
    return (csh);
}

static cpp_cshiislo cshiislo_lh_posun(cpp_cshiislo csh,
                                      size_t presnosth,
                                      size_t posun)
{
    if (posun >= presnosth)
    {
        csh.pretecshenie = !(csh.bezznamienkovee) && !cshiislo_je_nula(csh);
        csh.hornaa = csh.dolnaa = 0;
    }
    else
    {
        cpp_cshiislo originaal, spaeth;
        size_t posun2 = posun;

        originaal = csh;
        if (posun2 >= PRESNOSTH_CSHASTI)
        {
            posun2 -= PRESNOSTH_CSHASTI;
            csh.hornaa = csh.dolnaa;
            csh.dolnaa = 0;
        }
        if (posun2)
        {
            csh.hornaa = ((csh.hornaa << posun2)
                          | (csh.dolnaa >> (PRESNOSTH_CSHASTI - posun2)));
            csh.dolnaa <<= posun2;
        }
        csh = cshiislo_orezh(csh, presnosth);
        if (csh.bezznamienkovee) csh.pretecshenie = 0;
        else
        {
            spaeth = cshiislo_p_posun(csh, presnosth, posun);
            csh.pretecshenie = !cshiisla_suu_rovnee(originaal, spaeth);
        }
    }

    return (csh);
}

static cpp_cshiislo cshiislo_binaarna_op(cpp_cshii *cshii,
                                         cpp_cshiislo csh1,
                                         cpp_cshiislo csh2,
                                         enum cpp_typtokenu op)
{
    cpp_cshiislo vyysledok;
    size_t presnosth = CPP_NASTAVENIE(cshii, presnosth);

    switch (op)
    {
        case CPP_PLUS:
            vyysledok.dolnaa = csh1.dolnaa + csh2.dolnaa;
            vyysledok.hornaa = csh1.hornaa + csh2.hornaa;
            if (vyysledok.dolnaa < csh1.dolnaa)
                vyysledok.hornaa++;
            vyysledok.bezznamienkovee = (csh1.bezznamienkovee
                                        || (csh2.bezznamienkovee));
            vyysledok.pretecshenie = 0;

            vyysledok = cshiislo_orezh(vyysledok, presnosth);
            if (!(vyysledok.bezznamienkovee))
            {
                vyysledok.pretecshenie = ((cshiislo_kladnee(csh1, presnosth)
                                           == cshiislo_kladnee(csh2, presnosth))
                                          &&
                                          (cshiislo_kladnee(csh1, presnosth)
                                           != cshiislo_kladnee(vyysledok,
                                                               presnosth)));
            }
            return (vyysledok);

        case CPP_MIINUS:
            vyysledok.dolnaa = csh1.dolnaa - csh2.dolnaa;
            vyysledok.hornaa = csh1.hornaa - csh2.hornaa;
            if (vyysledok.dolnaa > csh1.dolnaa)
                vyysledok.hornaa--;
            vyysledok.bezznamienkovee = (csh1.bezznamienkovee
                                        || (csh2.bezznamienkovee));
            vyysledok.pretecshenie = 0;

            vyysledok = cshiislo_orezh(vyysledok, presnosth);
            if (!(vyysledok.bezznamienkovee))
            {
                vyysledok.pretecshenie = ((cshiislo_kladnee(csh1, presnosth)
                                           != cshiislo_kladnee(csh2, presnosth))
                                          &&
                                          (cshiislo_kladnee(csh1, presnosth)
                                           != cshiislo_kladnee(vyysledok,
                                                               presnosth)));
            }
            return (vyysledok);
        
        case CPP_LH_POSUN:
        case CPP_P_POSUN: {
            size_t posun;
            if (!(csh2.bezznamienkovee) && !cshiislo_kladnee(csh2, presnosth))
            {
                /* Zaapornyy posun jednyym smerom je kladnyy posun druhyym. */
                op = op == CPP_LH_POSUN ? CPP_P_POSUN : CPP_LH_POSUN;
                csh2 = cshiislo_zneguj(csh2, presnosth);
            }
            if (csh2.hornaa) posun = ~0;
            else posun = csh2.dolnaa;
            if (op == CPP_LH_POSUN)
                csh1 = cshiislo_lh_posun(csh1, presnosth, posun);
            else
                csh1 = cshiislo_p_posun(csh1, presnosth, posun);
            break;
        }

        case CPP_CSHIARKA:
            if (!(cshii->stav.preskocsh_vyhodnotenie))
                printf("cshiarkovyy operaator v preprocesorovom vyyraze\n");
            csh1 = csh2;
            break;

        default:
            printf("Nespraavnaa operaacia v cshiislo_binaarna_op\n");
            break;
    }

    return (csh1);
}

/* Vraati 1, ak csh1 >= csh2. */
static int cshiislo_vaecshie_rovnee(cpp_cshiislo csh1,
                                    cpp_cshiislo csh2,
                                    size_t presnosth)
{
    int z = csh1.bezznamienkovee || (csh2.bezznamienkovee);

    if (!z)
    {
        /* Obe cshiisla suu znamienkovee,
         * takzhe ak sa znamienka liishia,
         * vyysledok je znamienko prveeho. */
        z = cshiislo_kladnee(csh1, presnosth);

        if (z != cshiislo_kladnee(csh2, presnosth))
            return (z);

        /* Obe majuu rovnakee znamienko. */
    }

    return ((csh1.hornaa > csh2.hornaa)
            || ((csh1.hornaa == csh2.hornaa)
                && (csh1.dolnaa >= csh2.dolnaa)));
}

static cpp_cshiislo cshiislo_nerovnostnaa_op(cpp_cshii *cshii,
                                             cpp_cshiislo csh1,
                                             cpp_cshiislo csh2,
                                             enum cpp_typtokenu op)
{
    int vr = cshiislo_vaecshie_rovnee(csh1, csh2,
                                      CPP_NASTAVENIE(cshii, presnosth));

    switch (op)
    {
        case CPP_VAECSHIE:
            csh1.dolnaa = vr && !cshiisla_suu_rovnee(csh1, csh2);
            break;
        
        case CPP_MENSHIE:
            csh1.dolnaa = !vr;
            break;
        
        case CPP_VAECSHIE_ROVNAA_SA:
            csh1.dolnaa = vr;
            break;
        
        case CPP_MENSHIE_ROVNAA_SA:
            csh1.dolnaa = !vr || cshiisla_suu_rovnee(csh1, csh2);
            break;

        default:
            printf("Nespraavnaa operaacia v cshiislo_nerovnostnaa_op\n");
            csh1.dolnaa = 0;
            break;
    }
    
    csh1.hornaa = 0;
    csh1.bezznamienkovee = 0;
    csh1.pretecshenie = 0;
    return (csh1);
}

static cpp_cshiislo cshiislo_rovnostnaa_op(cpp_cshii *cshii,
                                           cpp_cshiislo csh1,
                                           cpp_cshiislo csh2,
                                           enum cpp_typtokenu op)
{
    switch (op)
    {
        case CPP_ROVNAA_SA_ROVNAA_SA:
            csh1.dolnaa = cshiisla_suu_rovnee(csh1, csh2);
            break;
        
        case CPP_NIE_ROVNAA_SA:
            csh1.dolnaa = !cshiisla_suu_rovnee(csh1, csh2);
            break;

        default:
            printf("Nespraavnaa operaacia v cshiislo_rovnostnaa_op\n");
            csh1.dolnaa = 0;
            break;
    }

    csh1.hornaa = 0;
    csh1.bezznamienkovee = 0;
    csh1.pretecshenie = 0;
    return (csh1);
}

static cpp_cshiislo cshiislo_bitovaa_op(cpp_cshii *cshii,
                                        cpp_cshiislo csh1,
                                        cpp_cshiislo csh2,
                                        enum cpp_typtokenu op)
{
    csh1.pretecshenie = 0;
    csh1.bezznamienkovee = csh1.bezznamienkovee || (csh2.bezznamienkovee);

    switch (op)
    {
        case CPP_A:
            csh1.hornaa &= csh2.hornaa;
            csh1.dolnaa &= csh2.dolnaa;
            break;
        
        case CPP_ALEBO:
            csh1.hornaa |= csh2.hornaa;
            csh1.dolnaa |= csh2.dolnaa;
            break;
        
        case CPP_XALEBO:
            csh1.hornaa ^= csh2.hornaa;
            csh1.dolnaa ^= csh2.dolnaa;
            break;

        default:
            printf("Nespraavnaa operaacia v cshiislo_bitovaa_op\n");
            csh1.dolnaa = 0;
            break;
    }

    return (csh1);
}

static cpp_cshiislo cshiislo_cshasth_naasob(cpp_cshasth_cshiisla csh1,
                                            cpp_cshasth_cshiisla csh2)
{
    cpp_cshiislo vyysledok;
    cpp_cshasth_cshiisla stred[2], docsh;

    vyysledok.dolnaa = DOLNAA_CSHASTH(csh1) * DOLNAA_CSHASTH(csh2);
    vyysledok.hornaa = HORNAA_CSHASTH(csh1) * HORNAA_CSHASTH(csh2);

    stred[0] = DOLNAA_CSHASTH(csh1) * HORNAA_CSHASTH(csh2);
    stred[1] = HORNAA_CSHASTH(csh1) * DOLNAA_CSHASTH(csh2);

    docsh = vyysledok.dolnaa;
    vyysledok.dolnaa += DOLNAA_CSHASTH(stred[0]) << (PRESNOSTH_CSHASTI / 2);
    if (vyysledok.dolnaa < docsh)
        vyysledok.hornaa++;

    docsh = vyysledok.dolnaa;
    vyysledok.dolnaa += DOLNAA_CSHASTH(stred[1]) << (PRESNOSTH_CSHASTI / 2);
    if (vyysledok.dolnaa < docsh)
        vyysledok.hornaa++;

    vyysledok.hornaa += HORNAA_CSHASTH(stred[0]);
    vyysledok.hornaa += HORNAA_CSHASTH(stred[1]);
    vyysledok.bezznamienkovee = 1;
    vyysledok.pretecshenie = 0;

    return (vyysledok);
}

static cpp_cshiislo cshiislo_naasob(cpp_cshii *cshii,
                                    cpp_cshiislo csh1,
                                    cpp_cshiislo csh2)
{
    int zneguj = 0;
    int bezznamienkovee = csh1.bezznamienkovee || (csh2.bezznamienkovee);
    int pretecshenie;
    size_t presnosth = CPP_NASTAVENIE(cshii, presnosth);
    cpp_cshiislo vyysledok, docsh;

    /* Priiprava na bezznamienkovee naasobenie. */
    if (!bezznamienkovee)
    {
        if (!cshiislo_kladnee(csh1, presnosth))
        {
            zneguj = !zneguj;
            csh1 = cshiislo_zneguj(csh1, presnosth);
        }
        if (!cshiislo_kladnee(csh2, presnosth))
        {
            zneguj = !zneguj;
            csh2 = cshiislo_zneguj(csh2, presnosth);
        }
    }

    pretecshenie = csh1.hornaa && csh2.hornaa;
    vyysledok = cshiislo_cshasth_naasob(csh1.dolnaa, csh2.dolnaa);

    docsh = cshiislo_cshasth_naasob(csh1.hornaa, csh2.dolnaa);
    vyysledok.hornaa += docsh.dolnaa;
    if (docsh.hornaa) pretecshenie = 1;

    docsh = cshiislo_cshasth_naasob(csh1.dolnaa, csh2.hornaa);
    vyysledok.hornaa += docsh.dolnaa;
    if (docsh.hornaa) pretecshenie = 1;

    docsh.dolnaa = vyysledok.dolnaa;
    docsh.hornaa = vyysledok.hornaa;
    vyysledok = cshiislo_orezh(vyysledok, presnosth);
    if (!cshiisla_suu_rovnee(vyysledok, docsh)) pretecshenie = 1;

    if (zneguj) vyysledok = cshiislo_zneguj(vyysledok, presnosth);

    if (bezznamienkovee) vyysledok.pretecshenie = 0;
    else
    {
        vyysledok.pretecshenie = (pretecshenie
                                  || ((cshiislo_kladnee(vyysledok, presnosth)
                                       ^ !zneguj)
                                      && !cshiislo_je_nula(vyysledok)));
    }

    vyysledok.bezznamienkovee = bezznamienkovee;

    return (vyysledok);
}

static cpp_cshiislo cshiislo_deliaca_op(cpp_cshii *cshii,
                                        cpp_cshiislo csh1,
                                        cpp_cshiislo csh2,
                                        enum cpp_typtokenu op)
{
    int zneguj = 0;
    int zneguj_csh1 = 0;
    int bezznamienkovee = csh1.bezznamienkovee || (csh2.bezznamienkovee);
    size_t i, presnosth = CPP_NASTAVENIE(cshii, presnosth);
    cpp_cshiislo vyysledok, odcshiitaj;
    cpp_cshasth_cshiisla maska;

    /* Priiprava na bezznamienkovee naasobenie. */
    if (!bezznamienkovee)
    {
        if (!cshiislo_kladnee(csh1, presnosth))
        {
            zneguj = !zneguj;
            zneguj_csh1 = 1;
            csh1 = cshiislo_zneguj(csh1, presnosth);
        }
        if (!cshiislo_kladnee(csh2, presnosth))
        {
            zneguj = !zneguj;
            csh2 = cshiislo_zneguj(csh2, presnosth);
        }
    }

    if (csh2.hornaa)
    {
        i = presnosth - 1;
        maska = ((cpp_cshasth_cshiisla)1) << (i - PRESNOSTH_CSHASTI);
        for (;; i--, maska >>= 1)
        {
            if (csh2.hornaa & maska) break;
        }
    }
    else if (csh2.dolnaa)
    {
        if (presnosth > PRESNOSTH_CSHASTI)
        {
            i = presnosth - PRESNOSTH_CSHASTI - 1;
        }
        else i = presnosth - 1;
        maska = ((cpp_cshasth_cshiisla)1) << i;
        for (;; i--, maska >>= 1)
        {
            if (csh2.dolnaa & maska) break;
        }
    }
    else
    {
        if (!(cshii->stav.preskocsh_vyhodnotenie))
            printf("delenie 0 v preprocesorovom vyyraze\n");
        return (csh1);
    }

    csh1.bezznamienkovee = 1;
    csh2.bezznamienkovee = 1;
    i = presnosth - i - 1;
    odcshiitaj = cshiislo_lh_posun(csh2, presnosth, i);

    vyysledok.hornaa = vyysledok.dolnaa = 0;
    for (;;)
    {
        if (cshiislo_vaecshie_rovnee(csh1, odcshiitaj, presnosth))
        {
            csh1 = cshiislo_binaarna_op(cshii, csh1, odcshiitaj, CPP_MIINUS);
            if (i >= PRESNOSTH_CSHASTI)
                vyysledok.hornaa |= (((cpp_cshasth_cshiisla)1)
                                     << (i - PRESNOSTH_CSHASTI));
            else
                vyysledok.dolnaa |= ((cpp_cshasth_cshiisla)1) << i;
        }
        if (i-- == 0) break;
        odcshiitaj.dolnaa = ((odcshiitaj.dolnaa >> 1)
                            || (odcshiitaj.hornaa << (PRESNOSTH_CSHASTI - 1)));
        odcshiitaj.hornaa >>= 1;
    }

    switch (op)
    {
        case CPP_LOMKA:
            vyysledok.bezznamienkovee = bezznamienkovee;
            vyysledok.pretecshenie = 0;
            if (!bezznamienkovee)
            {
                if (zneguj) vyysledok = cshiislo_zneguj(vyysledok, presnosth);
                vyysledok.pretecshenie = ((cshiislo_kladnee(vyysledok, presnosth)
                                           ^ !zneguj)
                                          && !cshiislo_je_nula(vyysledok));
            }

            return (vyysledok);

        case CPP_MODULO:
            csh1.bezznamienkovee = bezznamienkovee;
            csh1.pretecshenie = 0;
            if (zneguj_csh1) csh1 = cshiislo_zneguj(csh1, presnosth);
            break;
                
        default:
            printf("Nespraavnaa operaacia v cshiislo_deliaca_op\n");
            break;
    }

    return (csh1);
}

static struct op *redukuj(cpp_cshii *cshii,
                          struct op *vrch,
                          enum cpp_typtokenu op)
{
    unsigned int priorita;

    if ((vrch->op <= CPP_ROVNAA_SA) || (vrch->op > CPP_POSLEDNYY_CPP_OP + 2))
    {
        cpp_chyba(cshii, CPP_DUU_INTERNAA_CHYBA,
                 "nemozhnyy operaator '%u'", vrch->op);
        return (NULL);
    }

    if (op == CPP_LH_ZAATVORKA) return (vrch);

    priorita = (optab[op].priorita
                - (optab[op].vlajky & LHAVO_ASOCIATIIVNE != 0));
    while (priorita < optab[vrch->op].priorita)
    {
        if (optab[vrch->op].vlajky & SKONTROLUJ_POVYYSHENIE)
            skontroluj_povyyshenie(cshii, vrch);

        switch (vrch->op)
        {
            case CPP_UPLUS:
            case CPP_UMIINUS:
            case CPP_NIE:
            case CPP_KOMPLEMENT:
                vrch[-1].hodnota = cshiislo_unaarna_op(cshii,
                                                       vrch->hodnota,
                                                       vrch->op);
                break;

            case CPP_PLUS:
            case CPP_MIINUS:
            case CPP_LH_POSUN:
            case CPP_P_POSUN:
            case CPP_CSHIARKA:
                vrch[-1].hodnota = cshiislo_binaarna_op(cshii,
                                                        vrch[-1].hodnota,
                                                        vrch->hodnota,
                                                        vrch->op);
                break;

            case CPP_VAECSHIE:
            case CPP_MENSHIE:
            case CPP_VAECSHIE_ROVNAA_SA:
            case CPP_MENSHIE_ROVNAA_SA:
                vrch[-1].hodnota = cshiislo_nerovnostnaa_op(cshii,
                                                            vrch[-1].hodnota,
                                                            vrch->hodnota,
                                                            vrch->op);
                break;

            case CPP_ROVNAA_SA_ROVNAA_SA:
            case CPP_NIE_ROVNAA_SA:
                vrch[-1].hodnota = cshiislo_rovnostnaa_op(cshii,
                                                          vrch[-1].hodnota,
                                                          vrch->hodnota,
                                                          vrch->op);
                break;

            case CPP_A:
            case CPP_ALEBO:
            case CPP_XALEBO:
                vrch[-1].hodnota = cshiislo_bitovaa_op(cshii,
                                                       vrch[-1].hodnota,
                                                       vrch->hodnota,
                                                       vrch->op);
                break;

            case CPP_HVIEZDA:
                vrch[-1].hodnota = cshiislo_naasob(cshii,
                                                   vrch[-1].hodnota,
                                                   vrch->hodnota);
                break;

            case CPP_LOMKA:
            case CPP_MODULO:
                vrch[-1].hodnota = cshiislo_deliaca_op(cshii,
                                                       vrch[-1].hodnota,
                                                       vrch->hodnota,
                                                       vrch->op);
                break;

            case CPP_A_A:
                vrch--;
                if (cshiislo_je_nula(vrch->hodnota))
                    cshii->stav.preskocsh_vyhodnotenie--;
                vrch->hodnota.dolnaa = (!cshiislo_je_nula(vrch->hodnota)
                                        && !cshiislo_je_nula(vrch[1].hodnota));
                vrch->hodnota.hornaa = 0;
                vrch->hodnota.bezznamienkovee = 0;
                vrch->hodnota.pretecshenie = 0;
                continue;

            case CPP_ALEBO_ALEBO:
                vrch--;
                if (!cshiislo_je_nula(vrch->hodnota))
                    cshii->stav.preskocsh_vyhodnotenie--;
                vrch->hodnota.dolnaa = (!cshiislo_je_nula(vrch->hodnota)
                                        || !cshiislo_je_nula(vrch[1].hodnota));
                vrch->hodnota.hornaa = 0;
                vrch->hodnota.bezznamienkovee = 0;
                vrch->hodnota.pretecshenie = 0;
                continue;

            case CPP_LH_ZAATVORKA:
                if (op != CPP_P_ZAATVORKA)
                {
                    printf("chyyba ')' v preprocesorovom vyyraze\n");
                    return (NULL);
                }
                vrch--;
                vrch->hodnota = vrch[1].hodnota;
                return (vrch);

            case CPP_OTAAZNIK:
                if ((op == CPP_CSHIARKA) || (op == CPP_DVOJBODKA))
                    return (vrch);
                printf("'?' bez nasledujuucej ':'\n");
                return (NULL);

            case CPP_DVOJBODKA:
                vrch -= 2;
                if (!cshiislo_je_nula(vrch->hodnota))
                {
                    cshii->stav.preskocsh_vyhodnotenie--;
                    vrch->hodnota = vrch[1].hodnota;
                }
                else
                {
                    vrch->hodnota = vrch[2].hodnota;
                }
                vrch->hodnota.bezznamienkovee = ((vrch[1].hodnota
                                                  .bezznamienkovee)
                                                 || (vrch[2].hodnota
                                                     .bezznamienkovee));
                continue;

            default:
                cpp_chyba(cshii, CPP_DUU_INTERNAA_CHYBA,
                          "nemozhnyy operaator '%u'", vrch->op);
                return (NULL);
        }

        vrch--;
        if (vrch->hodnota.pretecshenie
            && !(cshii->stav.preskocsh_vyhodnotenie))
            cpp_chyba(cshii, CPP_DUU_VAROVANIE,
                      "pretecshenie celeeho cshiisla v preprocesorovom vyyraze");
    }

    if (op == CPP_P_ZAATVORKA)
    {
        cpp_chyba(cshii, CPP_DUU_CHYBA,
                  "chyyba '(' vo vyyraze");
        return (NULL);
    }
    
    return (vrch);
}

/* Spracuje vyyraz z cshii a vraati 1, ak je vyysledok nenulovyy,
 * inak 0. */
int _cpp_spracuj_vyyraz(cpp_cshii *cshii, int je_if)
{
    struct op *vrch = cshii->op_akumulaator;
    int chce_hodnotu = 1;
    
    cshii->stav.preskocsh_vyhodnotenie = 0;

    vrch->op = CPP_KONIEC;
    
    for (;;)
    {
        struct op op;

        op.token = cpp_daj_token(cshii);
        op.op = op.token->typ;

        switch (op.op)
        {
            /* Tieto sa zmenia na hodnoty. */
            case CPP_CSHIISLO:
            case CPP_ZNAK:
            case CPP_LZNAK:
            case CPP_uZNAK:
            case CPP_UZNAK:
            case CPP_IDENT:
                if (!chce_hodnotu)
                {
                    printf("chyyba binaarny operaator pred tokenom \"%s\"\n",
                           cpp_token_ako_text(op.token));
                    return (0);
                }
                chce_hodnotu = 0;
                vrch->hodnota = vyhodnoth_token(cshii, op.token);
                continue;

            case CPP_PLUS:
                if (chce_hodnotu) op.op = CPP_UPLUS;
                break;

            case CPP_MIINUS:
                if (chce_hodnotu) op.op = CPP_UMIINUS;
                break;

            default:
                if ((op.op <= CPP_ROVNAA_SA) || (op.op >= CPP_PLUS_ROVNAA_SA))
                {
                    printf("token \"%s\" nie je spraavny"
                           "v preprocesorovom vyyraze\n",
                           cpp_token_ako_text(op.token));
                    return (0);
                }
        }

        if (optab[op.op].vlajky & ZHIADNY_LH_OPERAND)
        {
            if (!chce_hodnotu)
            {
                printf("chyyba binaarny operaator pred tokenom \"%s\"\n",
                       cpp_token_ako_text(op.token));
                return (0);
            }
        }
        else if (chce_hodnotu)
        {
            if ((vrch->op == CPP_LH_ZAATVORKA) && (op.op == CPP_P_ZAATVORKA))
            {
                printf("chyyba vyyraz medzi '(' a ')'\n");
                return (0);
            }
            else if ((vrch->op == CPP_KONIEC) && (op.op == CPP_KONIEC))
            {
                printf("%s bez vyyrazu\n", je_if ? "#if" : "#elif");
                return (0);
            }
            else if ((vrch->op != CPP_KONIEC) && (vrch->op != CPP_LH_ZAATVORKA))
            {
                printf("operaator '%s' nemaa pravyy operand\n",
                       cpp_token_ako_text(vrch->token));
                return (0);
            }
            else if ((op.op == CPP_P_ZAATVORKA) || (op.op == CPP_KONIEC))
                /* Oznaami sa neskuor. */;
            else
            {
                printf("operaator '%s' nemaa lhavyy operand\n",
                       cpp_token_ako_text(op.token));
                return (0);
            }
        }
        
        vrch = redukuj(cshii, vrch, op.op);
        if (!vrch) return (0);

        if (op.op == CPP_KONIEC) break;

        switch (op.op)
        {
            case CPP_P_ZAATVORKA:
                continue;

            case CPP_ALEBO_ALEBO:
                if (!cshiislo_je_nula(vrch->hodnota))
                    cshii->stav.preskocsh_vyhodnotenie++;
                break;

            case CPP_A_A:
            case CPP_OTAAZNIK:
                if (cshiislo_je_nula(vrch->hodnota))
                    cshii->stav.preskocsh_vyhodnotenie++;
                break;

            case CPP_DVOJBODKA:
                if (vrch->op != CPP_OTAAZNIK)
                {
                    printf(" ':' bez predchaazdajuuceho '?'\n");
                    return (0);
                }
                if (!cshiislo_je_nula(vrch[-1].hodnota))
                    cshii->stav.preskocsh_vyhodnotenie++;
                else
                    cshii->stav.preskocsh_vyhodnotenie--;
                break;
        }

        chce_hodnotu = 1;

        if (++vrch == (cshii->op_koniec))
        {
            vrch = _cpp_expanduj_op_akumulaator(cshii);
        }

        vrch->op = op.op;
        vrch->token = op.token;
    }

    if (vrch != (cshii->op_akumulaator))
    {
        cpp_chyba(cshii, CPP_DUU_INTERNAA_CHYBA,
                  "nevyvaazhenyy akumulaator v %s",
                  je_if ? "#if" : "#elif");
        return (0);
    }
    
    return (!cshiislo_je_nula(vrch->hodnota));
}

struct op *_cpp_expanduj_op_akumulaator(cpp_cshii *cshii)
{
    size_t staraa_velhkosth = cshii->op_koniec - (cshii->op_akumulaator);
    size_t novaa_velhkosth = staraa_velhkosth * 2 + 20;

    cshii->op_akumulaator = realloc(cshii->op_akumulaator,
                                    sizeof(*(cshii->op_akumulaator)) * novaa_velhkosth);
    if (cshii->op_akumulaator == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    return (cshii->op_akumulaator + staraa_velhkosth);
}
