/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "cppkni.h"
#include "c_ppvyystup.h"
#include "include_cesta.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct odlozhenee_nastavenie {
    int kood;
    const char *arg;
};

typedef struct {
    cpp_cshii *cshii;
    include_cesty *ic;
    const char **menaa_vstupov;
    unsigned int pocshet_mien_vstupov;
    const char *meno_vyystupu;
    int vlajka_len_preprocesuj;
    void *spracovaavacsh;

    struct odlozhenee_nastavenie *odlozhenee_nastavenia;
    unsigned int odlozhenee_pocshet;

    int hosthovanee;
    
} zck_jadro;

zck_jadro *zck_vytvor_jadro(void)
{
    zck_jadro *zj = malloc(sizeof(*zj));

    if (zj == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    zj->cshii = NULL;
    zj->ic = NULL;
    zj->menaa_vstupov = NULL;
    zj->pocshet_mien_vstupov = 0;
    zj->meno_vyystupu = NULL;
    zj->vlajka_len_preprocesuj = 0;

    zj->odlozhenee_nastavenia = NULL;
    zj->odlozhenee_pocshet = 0;

    zj->hosthovanee = 1;

    return (zj);
}

void zck_znicsh_jadro(zck_jadro *zj)
{
    free(zj->menaa_vstupov);
    free(zj->odlozhenee_nastavenia);
    free(zj);
}

enum {
    NASTAVENIE_D
};

static void odlozh_nastavenie(zck_jadro *zj, int kood, const char *arg)
{
    zj->odlozhenee_nastavenia[zj->odlozhenee_pocshet].kood = kood;
    zj->odlozhenee_nastavenia[zj->odlozhenee_pocshet].arg = arg;
    zj->odlozhenee_pocshet++;
}

static void diagnostika(cpp_cshii *cshii, enum cpp_diagnostickaa_uurovenh uu,
                        lokaacia_t lok,
                        const char *spraava, va_list *vl)
{
    printf("\n");
    if (lok.suubor)
    {
        printf("%s:%i:%i: ", lok.suubor, lok.riadok, lok.stllpec);
    }
    switch(uu)
    {
        case CPP_DUU_VAROVANIE: printf("varovanie: "); break;
        case CPP_DUU_CHYBA: printf("chyba: "); break;
        case CPP_DUU_ZLYHANIE: printf("zlyhanie: "); break;
        case CPP_DUU_INTERNAA_CHYBA: printf("internaa chyba: "); break;
    }
    vprintf(spraava, *vl);
    printf("\n");
    if (uu == CPP_DUU_ZLYHANIE)
    {
        /* +++Zlyhanie kompilaacie */
    }
}

void c_inic_nastavenia(zck_jadro *zj, unsigned int pocshet_nastavenii,
                       void *nastavenia)
{
    zj->cshii = cpp_vytvor_cshii(SHTD_C17, NULL);
    zj->ic = ic_vytvor_include_cesty();

    cpp_daj_spaetneevolania(zj->cshii)->diagnostika = &diagnostika;

    zj->odlozhenee_nastavenia = malloc(sizeof(struct odlozhenee_nastavenie)
                                       * pocshet_nastavenii);
    if (zj->odlozhenee_nastavenia == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }
    zj->odlozhenee_pocshet = 0;
}

void c_po_nastavenia(zck_jadro *zj)
{
    ic_registruj_include_rethaze(zj->ic, zj->cshii);
    
    if (zj->vlajka_len_preprocesuj)
    {
        FILE *vyystup;

        if ((zj->meno_vyystupu == NULL)
            || (zj->meno_vyystupu[0] == '\0'))
        {
            vyystup = stdout;
        }
        else vyystup = fopen(zj->meno_vyystupu, "w");

        zj->spracovaavacsh = inic_pp_vyystup(zj->cshii, vyystup);
    }
    else
    {
        /* inic_c_lex() */
        printf("+++DOROB c_po_nastavenia\n");
    }

    if (cpp_cshiitaj_hlavnyy_suubor(zj->cshii, zj->menaa_vstupov[0]) == NULL)
    {
        printf("+++CHYBA c_po_nastavenia\n");
        abort();
    }
}

static void c_dokoncshi_nastavenia(zck_jadro *zj)
{
    lokaacia_t lok;
    unsigned int i;

    lok.suubor = "<vstavanee>";
    lok.riadok = 1;
    lok.stllpec = 0;
    cpp_daj_spaetneevolania(zj->cshii)->zmena_suuboru(zj->cshii, lok,
                                                      CPP_D_PREMENOVANIE);

    cpp_vynucuj_lokaaciu_tokenov(zj->cshii, lok);
    cpp_inic_vstavanee(zj->cshii, zj->hosthovanee);

    lok.suubor = "<priikazovyy-riadok>";
    lok.riadok = 1;
    lok.stllpec = 0;
    cpp_daj_spaetneevolania(zj->cshii)->zmena_suuboru(zj->cshii, lok,
                                                      CPP_D_PREMENOVANIE);

    cpp_vynucuj_lokaaciu_tokenov(zj->cshii, lok);

    for (i = 0; i < zj->odlozhenee_pocshet; i++)
    {
        struct odlozhenee_nastavenie *nastavenie;
        nastavenie = &(zj->odlozhenee_nastavenia[i]);

        switch (nastavenie->kood)
        {
            case NASTAVENIE_D:
                cpp_define(zj->cshii, nastavenie->arg);
                break;
        }
    }

    cpp_prestanh_vynucovath_lokaaciu_tokenov(zj->cshii);
}

static void c_koniec(zck_jadro *zj)
{
    cpp_znicsh_cshii(zj->cshii);
    free(zj->ic);
    if (zj->vlajka_len_preprocesuj)
    {
        free(zj->spracovaavacsh);
    }
}

static char *xrethdup(const char *reth)
{
    size_t dllzhka = strlen(reth);
    char *von = malloc(dllzhka + 1);
    if (von == NULL)
    {
        printf("Chyba pri alokaacii\n");
        abort();
    }

    strcpy(von, reth);

    return (von);
}

int main(int argc, char **argv)
{
    int i;
    
    zck_jadro *zj;

    for (i = 1; i < argc; i++)
    {
        if (strcmp(argv[i], "--pomoc") == 0)
        {
            printf("Pouzhitie: zck [nastavenia] suubor...\n");
            printf("Nastavenia:\n");
            printf("  --pomoc        Zobrazii tieto informaacie\n");
            printf("  -E             Len preprocesuje; nekompiluje\n");
            printf("  -o <suubor>    Vlozhii vyystup do <suuboru>\n");
            printf("  -I <adresaar>  #include hladaa aj v <adresaari>\n");

            return (0);
        }
    }

    zj = zck_vytvor_jadro();

    c_inic_nastavenia(zj, argc - 1, NULL);

    for (i = 1; i < argc; i++)
    {
        if (argv[i][0] == '-')
        {
            argv[i]++;
            switch (argv[i][0])
            {
                case 'E':
                    zj->vlajka_len_preprocesuj = 1;
                    break;

                case 'I':
                    if (argv[i][1] != '\0')
                    {
                        ic_pridaj_cestu(zj->ic, xrethdup(argv[i] + 1),
                                        INCLUDE_CESTA_SHIKMEE);
                    }
                    else
                    {
                        ic_pridaj_cestu(zj->ic, xrethdup(argv[++i]),
                                        INCLUDE_CESTA_SHIKMEE);
                    }
                    break;

                case 'o':
                    if (argv[i][1] != '\0')
                    {
                        zj->meno_vyystupu = argv[i] + 1;
                    }
                    else
                    {
                        zj->meno_vyystupu = argv[++i];
                    }
                
                    break;

                case 'D':
                    if (argv[i][1] != '\0')
                    {
                        odlozh_nastavenie(zj, NASTAVENIE_D, argv[i] + 1);
                    }
                    else
                    {
                        odlozh_nastavenie(zj, NASTAVENIE_D, argv[++i]);
                    }
                
                    break;
            }
        }
        else
        {
            zj->pocshet_mien_vstupov++;
            zj->menaa_vstupov = realloc(zj->menaa_vstupov,
                                        (sizeof(*(zj->menaa_vstupov))
                                         * (zj->pocshet_mien_vstupov)));
            if (zj->menaa_vstupov == NULL)
            {
                printf("Chyba pri alokaacii\n");
                abort();
            }
            zj->menaa_vstupov[zj->pocshet_mien_vstupov - 1] = argv[i];
        }
    }

    if (zj->pocshet_mien_vstupov == 0)
    {
        printf("chyba: zhiadne vstupnee subory\n");
        goto koniec;
    }

    if (!(zj->vlajka_len_preprocesuj))
    {
        printf("+++s vyynimkou preprocesara kompilaator nie je hotovyy\n");
        printf("pouzhite moznosth -E\n");
        goto koniec;
    }
    
    c_po_nastavenia(zj);

    /* Zavolath pre kazdyy suubor. */
    c_dokoncshi_nastavenia(zj);

    if (zj->vlajka_len_preprocesuj)
    {
        preprocesuj_suubor(zj->spracovaavacsh, zj->cshii);
    }

koniec:
    c_koniec(zj);
    
    zck_znicsh_jadro(zj);

    return (0);
}
